# -*- coding: utf-8 -*-

from ..httppool.session import requests_retry_session
from ..dataset.product import Product, BaseProduct
from ..dataset.classes import Class_Look_Up
from ..dataset.serializable import serialize
from .poolmanager import PoolManager
from .productpool import ManagedPool, populate_pool2, ProductPool
from .productref import ProductRef
from .productstorage import ProductStorage
from .dicthk import HKDBS
from .urn import makeUrn, parse_poolurl, Urn, parseUrn
from ..pns.public_fdi_requests import read_from_cloud, load_from_cloud, delete_from_server
from ..pns.fdi_requests import ServerError
from ..utils.common import (fullname, lls, trbk,
                            logging_ERROR,
                            logging_WARNING,
                            logging_INFO,
                            logging_DEBUG
                            )

from ..utils.getconfig import getConfig

import logging
import os
from itertools import chain
import sys

logger = logging.getLogger(__name__)
pcc = getConfig()

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
else:
    PY3 = False
    strset = (str, unicode)

"""
Cloud apis classification:
#/%E9%85%8D%E7%BD%AE%E7%AE%A1%E7%90%86
DOC: http://123.56.102.90:31702/api/swagger-ui.html
Problem:
1. No class shown in storage/info API
2. No pattern for pool path, API use poolname+random name instead of pool <scheme>://<place><poolpath>/<poolname>
3. getMetaByUrn(self, urn, resourcetype=None, index=None) What's means resourcetype and index?
"""


def get_Values_From_A_list_of_dicts(res, what, get_list=True, excpt=True):
    """"Returns value(s) of a list of dicts containing key:value.

       If `get_list` is `True` always retur list of
     vals. Missing key returns `None` as val.
     `None` or Empty list gets `None`.
       If `get_list` is `False` try to return the  value of `what`
      in the first element of `res`. Return `None` upon any exception.
    :except: if set (default), throw `KeyError` if the product or `what` key is missing, or 0 is result.
    """
    if not get_list:
        try:
            return res[0]['what']
        except KeyError as e:
            if excpt:
                raise
            else:
                return 0
    if excpt:
        return [r.get(what, None) for r in res]
    else:
        if not res:
            return []
        return [r[what] for r in res]


class PublicClientPool(ManagedPool):
    def __init__(self,  auth=None, client=None, **kwds):
        """ creates file structure if there isn't one. if there is, read and populate house-keeping records. create persistent files if not exist.

        Parameters
        ----------
        auth : tuple, HTTPBasicAuth, or Authorization
            Authorization for remote pool.
        client : Request or Wrapper
            Mainly used for testing with mock server.
        **kwds :

        Returns
        -------

        """
        # print(__name__ + str(kwds))
        super().__init__(**kwds)
        self.auth = auth
        if client is None:
            client = requests_retry_session()
        self.client = client
        self.getToken()
        self.poolInfo = None

    def setup(self):
        """ Sets up HttpPool interals.

        Make sure that self._poolname and self._poolurl are present.
        """

        if super().setup():
            return True

        return False

    def setPoolurl(self, poolurl):
        """ Replaces the current poolurl of this pool.
            For cloud pool, there are also self._cloudpoolpath and self._cloudpoolname
            csdb:///csdb_test_pool
            self._poolpath, self._scheme, self._poolname = '', 'csdb', 'csdb_test_pool'
            self._cloudpoolpath = /csdb_test_pool
        """
        s = (not hasattr(self, '_poolurl') or not self._poolurl)
        self._poolpath, self._scheme, self._place, \
            self._poolname, self._username, self._password = \
            parse_poolurl(poolurl)
        if self._scheme == '' or self._scheme == None:
            self._scheme = 'csdb'
        self._cloudpoolpath = self._poolpath + '/' + self._poolname
        self._poolurl = poolurl
        # call setup only if poolurl was None
        if s:
            self.setup()

    def getPoolpath(self):
        if self._cloudpoolpath:
            return self._cloudpoolpath
        else:
            return self._poolpath

    def getToken(self):
        """Get CSDB acces token.

        Returns
        -------
        str
            Saved or newly gotten token.

        Raises
        ------
        RuntimeError
            Could not get new token.

        """
        err = False
        tokenMsg = None
        self.token = pcc['cloud_token']
        try:
            tokenMsg = read_from_cloud(
                'verifyToken', token=self.token, client=self.client)
        except ServerError as e:
            err = e
            pass
        if (tokenMsg is not None) and tokenMsg['status'] == 500 or err:
            logger.debug('Cloud token %s to be updated.' % self.token[:5])
        else:
            raise

        tokenMsg = read_from_cloud('getToken', client=self.client)
        if tokenMsg:
            self.token = tokenMsg['token']
            logger.debug('Cloud token %s updated.' % self.token[:5])
        else:
            raise ServerError('Update cloud token failed: ' +
                              tokenMsg['msg'])
        return self.token

    def poolExists(self, poolname=None):
        if poolname is None:
            poolname = self.poolname
        try:
            res = read_from_cloud(
                'existPool', poolname=poolname, token=self.token)
            return True
        except:
            return False

    def restorePool(self):
        res = read_from_cloud(
            'restorePool', poolname=self.poolname, token=self.token)
        if res['msg'] == 'success':
            return True
        else:
            return False

    def createPool2(self):
        try:
            res = read_from_cloud(
                'createPool', poolname=self.poolname, token=self.token)
        except:
            pass
        return None

    def createPool(self):
        ok, res = self.createPool2()
        return ok

    def getPoolInfo(self, update_hk=True, count=False):
        """
        Download, store server storage info.

        PoolInfo schema 1.0 consisted of three maps: classes,
        urns and tags data:. See productpool::ManagedPool::saveOne
            poolname :
                _classes= {dataTypeName: {currentSn:csn, sn=[]}}
                _urns= [{urn: tags[]}]
                _tags= {urns:[]}

        Parameters
        ----------
        update_hk : bool
            If set (default), get full PoolInfo from server and update local HK tables. If not set the result depends on `count`.
        count : bool
            Return count oriented results. Default is False.

        Results
        -------
        dict

            | update_hk | count  |  output     |
            | `True`    | any | Get the updated PoolInfo and save and returns updated PoolInfo |
            | `False`   | `True`  | Get and returns Count-format PoolInfo from `storage/info?getCount=1` |
            | `False`   | `False` | Returns the local PoolInfo in memory. |
        """
        if update_hk:
            res = read_from_cloud(
                'infoPool', pools=self.poolname, getCount=0, token=self.token)
            rdata = res
            if rdata:
                if self._poolname in rdata:
                    dpu = rdata[self._poolname]['_urns']
                    if count:
                        return len(dpu)
                    for urn, tags in dpu.items():
                        pool, ptype, sn = parseUrn(urn)
                        self._dTypes, self._dTags, sn = \
                            populate_pool2(tags, ptype, sn, self._dTypes,
                                           self._dTags)
                    logger.debug(
                        f'HK updated for the pool {self.poolname}.')
                else:
                    logger.info(
                        f'No pool found by the name {self.poolname}')
                self.poolInfo = rdata
                return rdata
            else:
                logger.info(f'No pool found.')
            return None
        else:
            # update_hk is False
            if count:
                res = read_from_cloud(
                    'infoPool', pools=self.poolname, getCount=1, token=self.token)
                return res
            else:
                return self.poolInfo

    def readHK(self):
        """ read housekeeping data from server """

        self.getPoolInfo(update_hk=True)

        return dict((n, db) for n, db in zip(HKDBS, [
            self._classes, self._tags, self._urns,
            self._dTypes, self._dTags]))

    def exists(self, urn, update=True):
        """
        Determines the existence of a product with specified URN.

        :update: read the server using `getPoolInfo` w/o updating the HK tbles. Defalut is `False`: only look up in the local cache of `PoolInfo`.
        """

        if update:
            self.getPoolInfo(update_hk=False)
        return urn in self.poolInfo[self._poolname]['_urns']

    def getProductClasses(self, count=True):
        """
        Returns all Product classes found in this pool.

            Ref. doc of `getPoolInfo`.

        Parameters
        -----------
        count : bool

        Returns
        -------
        list:
            of all product types.
        None:
            No such pool.
        """

        if count:
            cnt = self.getPoolInfo(update_hk=False, count=count)
            return list(cnt[self.poolname]['_classes'])
        if self.poolInfo is None:
            return None
        classes = list(self.poolInfo[self.poolname]['_classes'])
        return classes

    def getCount(self, typename=None, update=True):
        """
        Return the number of URNs for the product type in the pool.

        Parameters
        -----------
        update : bool
             Read the latest from server if set (default) or use the local `poolInfo`.
        typename : str, None
            Name of datatype. If is `None` get count of all types.

        Returns
        -------
        int:
            of all product types.
        None:
            No such pool.
        """
        cnt = self.getPoolInfo(update_hk=False, count=update)
        if update:
            # return cnt[self._poolname]['cnt'] if typename is None else cnt[self._poolname]['_classes'].get(typename, 0)
            cp = cnt.get(self._poolname, 0)
            if cp == 0:
                return 0
            return cp.get('_urns', 0) if typename is None else cp['_classes'].get(typename, {'cnt': 0})['cnt']
        return super().getCount(typename)

    def isEmpty(self):
        """
        Determines if the pool is empty.
        """
        res = self.getPoolInfo()
        if issubclass(res.__class__, dict):
            clses = res[self.poolname]['_classes']
            return len(clses) == 0
        else:
            raise ValueError('Error getting PoolInfo ' + str(res))

    def getMetaByUrn(self, urn=None, resourcetype=None, index=None):
        """
        Get all of the meta data belonging to a product of a given URN.

        mh: returns an iterator.
        """
        res = read_from_cloud(requestName='getMeta', urn=urn, token=self.token)
        return res

    def getDataType(self, urn=None):
        """Returns the Datatype of one or a list of URNs.

        Parameters
        ----------
        urn : str, list, None

        Returns
        -------
        dict, list
            all types if `urn` is not given. one or a list of
            datatypes (None if the URN is not found, '' of no
            Datatype.

        Examples
        --------
        FIXME: Add docs.

        """

        if urn is None:
            res = read_from_cloud('getDataType', token=self.token)
            return res

        if isinstance(urn, list):
            alist = True
        else:
            alist = False
            urns = [urn]
        paths = [u[3:].replace(':', '/') for u in urn]
        r = self.get_DataInfo('dataType', paths=paths)
        return r if alist else r[0]

    def getDataInfo(self, what='', paths=None, pool=None,
                    nulltype=True, limit=10000, asyn=False,
                    excpt=True):
        """Returns the CSDB storage information of one or a list of URNs.

        Parameters
        ----------
        what : str
            which item in list to return. e.g. 'urn' means
            a list
            of URNs found in the path. Default '' for all items.
        paths : str, list, None
            part in a path. URNs are allowed. Typical uses: '{poolname}' for a pool,
            or '/{poolname}/{product-type}' for a type of products in a pool,
            or '{poolname}/{product-name}/{index.aka.serial-number}',
            e.g. '/sv1/sv.BaseProduct' for all 'sv.BaseProduct' products in pool 'sv1'.
            Only one  of `paths` and `pool` can be non-empty.
            default is `None`.
        pool : str, None
            Only one  of `paths` and `pool` can be non-empty.
            Default is `None` for `self` pool if `paths` is empty,
            empty if `paths` is not.
        nulltype : bool
            Include None dataType entries. default `True`
        limit : int
            Maximum record number. Default `None` for 10000.
        except: bool
            If set (default), throw `KeyError` if the product or
        `what` key is missing, or 0 is result.
        Returns
        -------
        dict, list
            `None` if `paths` and `pool` are not found. one or a list of
            value for the given `what` key (None if not found).

        Examples
        --------
        FIXME: Add docs.

        """
        pname = self._poolname
        alist = isinstance(paths, (list, tuple))
        if not alist:
            if paths and pool:
                raise ValueError(
                    f"Path '{a}' and pool '{pool}' cannot be both non-empty for getDataInfo.")
            # if both are empty, pool takes self.poolname
            if not pool and not paths:
                pool = pname
            res = read_from_cloud('getDataInfo', token=self.token,
                                  paths=paths, pool=pool, limit=limit)
            # if input is not list the output of each query is not a list
            if not nulltype:
                popped = [res.pop(i) for i in range(
                    len(res)-1, -1, -1) if res[i]['dataType'] is None]
                logger.debug(f'{len(popped)} popped')
            if what:
                r = get_Values_From_A_list_of_dicts(res, what, excpt=excpt)
            else:
                # if path is a loner return the value, else
                # the pool is returned (one element pool, too).
                r = res
            return r
        # alist
        if 0:
            if a and pool:
                raise ValueError(
                    f"Path '{a}' and pool '{pool}' cannot be both non-empty for getDataInfo.")
        # if both are empty, pool takes self.poolname
        if not pool:
            pool = pname
        res = read_from_cloud('getDataInfo', token=self.token,
                              paths=paths, pool=pool, limit=limit, asyn=asyn)
        # in the output each query is a list[[p,p..],[p,p,...]]
        if not nulltype:
            popped = [q.pop(i) for q in res for i in range(
                len(q)-1, -1, -1) if q[i]['dataType'] is None]
            logger.debug(f'{len(popped)} popped')
        if what:
            bunch = [get_Values_From_A_list_of_dicts(r, what, excpt=excpt)
                     for r in res]
        else:
            bunch = res
        return bunch

    def doSave(self, resourcetype, index, data, tag=None, serialize_in=True, **kwds):
        path = f'/{self._poolname}/{resourcetype}'

        res = load_from_cloud('uploadProduct', token=self.token,
                              products=data, path=path, tags=tag, resourcetype=resourcetype, **kwds)
        return res

    def saveOne(self, prd, tag, geturnobjs, serialize_in, serialize_out, res, check_type=True, **kwds):
        """Save one product.

        Parameters
        ----------
        prd : BaseProduct, str, list
            The product(s) to be saved. It can be a product, a list of
            product or, when `serialize_in` is false, a JSON
            serialized product or list of products.
        tag : string list
             One or a list of strings. Comma is used to separate
             multiple tags in one string.
        geturnobjs : bool
            return URN object(s) instead of ProductRef(s).
        serialize_in : bool
            The input needs to be serialized for saving to a pool.
        serialize_out : bool
            The output is serialized.
        res : list
            the output list when input `prd` is a list.
        check_type : list
            To Check if the remote (CSDB) server has the type of the
            products that are to be saved.
        **kwds :

        Returns
        -------
        ProductRef, URN, list
            One ProductRef or URN object or a list of them. If
            `serialize_in` is true, the output PRoductRef has metadata
            in it.

        Raises
        ------
        ValueError
            Product type not found on the server.

        Examples
        --------
        FIXME: Add docs.


        """

        jsonPrd = prd
        if serialize_in:
            pn = fullname(prd)
            cls = prd.__class__
            jsonPrd = serialize(prd)
        else:
            # prd is json. extract prod name
            # '... "_STID": "Product"}]'
            pn = prd.rsplit('"', 2)[1]
            cls = Class_Look_Up[pn]
            pn = fullname(cls)

        if check_type:
            if pn not in check_type:
                raise ValueError('No such product type in csdb: ' + pn)

        # targetPoolpath = self.getPoolpath() + '/' + pn
        try:
            # save prod to cloud
            if serialize_in:
                uploadRes = self.doSave(resourcetype=pn,
                                        index=None,
                                        data=jsonPrd,
                                        tag=tag,
                                        serialize_in=serialize_in,
                                        serialize_out=serialize_out,
                                        **kwds)

            else:
                uploadRes = self.doSave(resourcetype=pn,
                                        index=None,
                                        data=prd,
                                        tag=tag,
                                        serialize_in=serialize_in,
                                        serialize_out=serialize_out,
                                        **kwds)
        except ValueError as e:
            msg = f'product {self.poolname}/{pn} saving failed. {e} {trbk(e)}'
            logger.debug(msg)
            raise e

        self._format_res(uploadRes, geturnobjs, prd, serialize_out, res)

    def _format_res(self, uploadRes, geturnobjs, prd, serialize_out, res):
        """ uploadRes : record of the csdb upload. """
        utype = uploadRes.get('dataType', '')
        if utype == '':
            raise Exception('Upload failed: product "%s" to %s:' % (
                prd.description, self.poolurl) + uploadRes['msg'])

        urn = uploadRes['urn']

        if utype is None:
            logger.warning(
                f'{urn} is uploaded but has no dataType. Upload Datatype definition to fix.')

        if geturnobjs:
            if serialize_out:
                # return the URN string.
                res.append(urn)
            else:
                res.append(Urn(urn))
        else:
            rf = ProductRef(urn=Urn(urn, poolurl=self.poolurl))
            if serialize_out:
                # return without meta
                res.append(rf)
            else:
                # it seems that there is no better way to set meta
                rf._meta = prd.getMeta() if getattr(prd, 'getMeta') else ''
                res.append(rf)

    def asyncSave(self, prds, tags, geturnobjs, serialize_in, serialize_out, res, check_type=[], **kwds):
        jsonPrds = []
        paths = []
        resourcetypes = []
        p0 = f'/{self._poolname}/%s'
        for prd in prds:
            if serialize_in:
                fc = fullname(prd)
                resourcetypes.append(fc)
                jsonPrds.append(serialize(prd))
            else:
                # prd is json. extract prod name
                # '... "_STID": "Product"}]'
                pn = prd.rsplit('"', 2)[1]
                cls = Class_Look_Up[pn]
                fc = fullname(cls)
                resourcetype.append(fc)
                jsonPrds.append(prd)
            paths.append(p0 % fc)

        if check_type:
            for pn in set(resourcetypes):
                if pn not in check_type:
                    raise ValueError('No such product type in cloud: ' + pn)
        # save prods to cloud
        if serialize_in:
            uploadr = load_from_cloud('uploadProduct', token=self.token,
                                      products=jsonPrds, path=paths,
                                      resourcetype=resourcetypes,
                                      tags=tags, asyn=True,
                                      **kwds)
            for i, uploadRes in enumerate(uploadr):
                self._format_res(uploadRes, geturnobjs,
                                 prd, serialize_out, res)
        return res

    def schematicSave(self, products, tag=None, geturnobjs=False, serialize_in=True, serialize_out=False, asyn=False, **kwds):
        """ do the scheme-specific saving.

            :serialize_out: if True returns contents in serialized form.
        """
        res = []
        if not PoolManager.isLoaded(self._poolname):  # self.poolExists():
            raise ValueError(f'Pool {self._poolname} is not registered.')

        check_type = self.getDataType()

        res = super().schematicSave(products, tag=tag, geturnobjs=geturnobjs,
                                    serialize_in=serialize_in,
                                    serialize_out=serialize_out,
                                    asyn=asyn,
                                    check_type=check_type,
                                    **kwds)
        # XXX refresh currentSn on server
        self.getPoolInfo()

        return res

    def schematicLoad(self, resourcetype, index, start=None, end=None,
                      serialize_out=False):
        """ do the scheme-specific loading
        """

        spn = self._poolname
        poolInfo = self.getPoolInfo()
        try:
            if poolInfo:
                pinfo = poolInfo
                if spn in pinfo:
                    if index in pinfo[spn]['_classes'][resourcetype]['sn']:
                        urn = makeUrn(poolname=spn,
                                      typename=resourcetype, index=index)
                        res = load_from_cloud(
                            'pullProduct', token=self.token, urn=urn)
                        # res is a product like ..dataset.product.Product

                        if issubclass(res.__class__, BaseProduct):
                            if serialize_out:
                                from ..dataset.deserialize import serialize
                                return serialize(res)
                            else:
                                return res
                        else:
                            raise Exception('Load failed: ' +
                                            res.get('msg', lls(res, 999)))
        except Exception as e:
            logger.debug('Load product failed:' + str(e))
            raise e
        logger.debug('No such product:' + resourcetype +
                     ' with index: ' + str(index))
        raise ValueError('No such product:' + resourcetype +
                         ' with index: ' + str(index))

    def doLoad(self, resourcetype, index, start=None, end=None, serialize_out=False):
        """ to be implemented by subclasses to do the action of loading
        """
        raise NotImplementedError

    # def schematicRemove(self, urn=None, resourcetype=None, index=None, asyn=False, **kwds):
    #     """ do the scheme-specific removing.

    #     `resourcetype+index` takes priority over `urn`.
    #     """
    #     prod, sn, urn = self.get_missing(urn, resourcetype, index,
    #                                      no_check=True)
    #     if index is None or index == '':
    #         self.schematicWipe()
    #         return

    #     if issubclass(index.__class__, (list, tuple)):
    #         if asyn:
    #             return
    #         else:
    #             for i, ind in enumerate(index):
    #                 self.doRemove(resourcetype, ind)
    #     else:
    #         # a single product
    #         return self.doRemove(resourcetype, index)

    def doRemove(self, resourcetype, index, asyn=False):
        """ to be implemented by subclasses to do the action of reemoving
        """

        path0 = f'/{self._poolname}'
        datatype, sns, alist = ProductPool.vectorize(resourcetype, index)

        path = [f'{path0}/{f}/{i}' for f, i in zip(datatype, sns)]
        res = read_from_cloud('remove', token=self.token, path=path, asyn=asyn)
        return res

    def XdoAsyncRemove(self, resourcetype, index):
        """ implemented with aio to do the action of removing asynchronously.
        """
        # path = self._cloudpoolpath + '/' + resourcetype + '/' + str(index)

        res = read_from_cloud('remove', token=self.token, path=path)
        if res['code'] and not getattr(self, 'ignore_error_when_delete', False):
            raise ValueError(
                f"Remove product {path} failed: {res['msg']}")
        logger.debug(f"Remove {path} code: {res['code']} {res['msg']}")

        return res

#     def schematicWipe(self, restore_type=True):
#
#         self.doWipe(restore_type)

    def doWipe(self, restore_type=False):
        """ to be implemented by subclasses to do the action of wiping.
        """
        poolname = self._poolname

        # res = read_from_cloud(
        #     'wipePool', poolname=poolname, token=self.token)
        # if res['msg'] != 'success':
        #     raise ValueError('Wipe pool ' + poolname +
        #                      ' failed: ' + res['msg'])
        # return
        cnt = self.getProductClasses(count=True)
        path = None
        res = []

        for clazz in cnt:
            path = f'/{poolname}/{clazz}'
            try:
                r = read_from_cloud(
                    'delDataTypeData', path=path, token=self.token)
            except Exception as e:
                msg = f'Wipe pool {poolname} failed: ' + str(e)
                if getattr(self, 'ignore_error_when_delete', False):
                    logger.warning(msg)
                else:
                    raise ValueError(msg)
            res.append(r)
            logger.debug(f'Removed {path}/{clazz}')
        logger.debug(f'Done removing {path}')
        return res

    def setTag(self, tag, urn):
        """ Set given tag or list of tags to the URN.

        Parameters
        ----------
        :tag: tag or list of tags.
        """
        u = urn.urn if issubclass(urn.__class__, Urn) else urn
        if not self.exists(urn):
            raise ValueError('Urn does not exist!')
        if isinstance(tag, (list, str)) and len(tag) > 0:
            # no space in "."
            t = ','.join(tag) if isinstance(tag, list) else tag
            try:
                res = read_from_cloud(
                    'addTag', token=self.token, tag=t, urn=u)
            except ServerError as e:
                msg = f'Set {tag} to {urn} failed: {e}'
                logger.warning(msg)
                raise
            except ValueError as e:
                msg = 'Tag can not be empty or non-string!'
                raise

    def getTags(self, urn=None, update=False):
        """

        :update: read the server using `getPoolInfo` w/o updating the HK tables. Default is `False`: only look up in the local cache of `PoolInfo`.
        """

        if update:
            self.getPoolInfo()
        ts = self.poolInfo[self._poolname]['_urns'].get(urn)
        return ts

    def removeTagByUrn(self, tag, urn):
        pass

    def removeTag(self, tag):
        if isinstance(tag, str):
            res = delete_from_server('delTag', token=self.token, tag=tag)
        else:
            raise ValueError('Tag must be a string!')

    def tagExists(self, tag):
        """
        Tests if a tag exists.


        Parameters
        ----------
        tag : str, list

        Returns
        -------
        dict, list
            Throw ValueError if `tag` is not given. one or a list of
            results (None if the tag is not found.

        Examples
        --------
        FIXME: Add docs.

        """

        if tag is None:
            raise ValueError('Cannot take None or "" as a tag.')

        if isinstance(tag, list):
            alist = True
        else:
            alist = False
            tags = [tag]
        try:
            res = read_from_cloud(
                'tagExist', token=self.token, tag=tag)
        except ServerError as e:
            msg = f'Get existancw tag failed: {e}'
            logger.warning(msg)
            raise
        return res

    def getUrn(self, tag=None):
        """
        Gets the URNs corresponding to the given tag.


        Parameters
        ----------
        tag : str, list

        Returns
        -------
        dict, list
            all URNs if `tag` is not given. one or a list of
            URNs (None if the tag is not found.

        Examples
        --------
        FIXME: Add docs.

        """

        if tag is None:
            raise ValueError('Cannot take None or "" as a tag.')

        if isinstance(tag, list):
            alist = True
        else:
            alist = False
            tags = [tag]
        try:
            res = read_from_cloud(
                'tagExist', token=self.token, tag=tags)
        except ServerError as e:
            msg = f'Get URNs by tag failed: {e}'
            logger.warning(msg)
            raise
        return res

    def meta_filter(self, q, typename=None, reflist=None, urnlist=None, snlist=None):
        """ returns filtered collection using the query.

        q is a MetaQuery
        valid inputs: typename and ns list; productref list; urn list
        """
        pass

    def prod_filter(self, q, cls=None, reflist=None, urnlist=None, snlist=None):
        """ returns filtered collection using the query.

        q: an AbstractQuery.
        valid inputs: cls and ns list; productref list; urn list
        """

    def doSelect(self, query, results=None):
        """
        to be implemented by subclasses to do the action of querying.
        """
        raise (NotImplementedError)


# =================SAVE REMOVE LOAD================
# test_getToken2()
# test_poolInfo()
# test_upload()
# test_get()
# test_remove()
# test_multi_upload()
# prd = genProduct(1)
# res = cp.schematicSave(prd)
# cp.schematicRemove('urn:poolbs:20211018:4')

# cp.schematicLoad('fdi.dataset.product.Product', 1)
# cp.schematicLoad('20211018', 5)



from fdi.dataset.product import _Model_Spec as PPI
from .product import Product
from .baseproduct import BaseProduct
from .numericparameter import NumericParameter
from .dateparameter import DateParameter
from .stringparameter import StringParameter
from .datatypes import Vector
from .dataset import CompositeDataset
from .tabledataset import TableDataset
from .arraydataset import ArrayDataset, Column
from ..pal.context import Context, MapContext
from ..pal.productref import ProductRef
from ..utils.loadfiles import loadMedia
from .finetime import FineTime

import copy
from math import sin, cos, sqrt
import random
from os import path as op


class TB(BaseProduct):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.zInfo['name'] = 'TB'
        self.zInfo['description'] = 'Test class %s.' % self.zInfo['name']
        self.zInfo['metadata']['type']['default'] = self.zInfo['name']


class TP(Product):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.zInfo['name'] = 'TP'
        self.zInfo['description'] = 'Test class %s.' % self.zInfo['name']
        self.zInfo['metadata']['type']['default'] = self.zInfo['name']


class TP_0X(TP):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.zInfo['name'] = 'TP_0X'
        self.zInfo['description'] = 'Test class %s.' % self.zInfo['name']
        self.zInfo['metadata']['type']['default'] = self.zInfo['name']


class TC(Context):
    def __init__(self, *args, **kwds):

        super().__init__(*args, **kwds)
        self.zInfo['name'] = 'TC'
        self.zInfo['description'] = 'Test class %s.' % self.zInfo['name']
        self.zInfo['metadata']['type']['default'] = self.zInfo['name']


class TCC(TC):
    def __init__(self, *args, **kwds):

        super().__init__(*args, **kwds)
        self.zInfo['name'] = 'TCC'
        self.zInfo['description'] = 'Test class %s.' % self.zInfo['name']
        self.zInfo['metadata']['type']['default'] = self.zInfo['name']


class TM(MapContext):
    def __init__(self, *args, **kwds):

        super().__init__(*args, **kwds)
        self.zInfo['name'] = 'TM'
        self.zInfo['description'] = 'Test class %s.' % self.zInfo['name']
        self.zInfo['metadata']['type']['default'] = self.zInfo['name']

# sub-classing testing class
# 'version' of subclass is int, not string


sp = copy.deepcopy(PPI)
sp['name'] = 'SP'
sp['metadata']['version']['data_type'] = 'integer'
sp['metadata']['version']['default'] = 9
sp['metadata']['type']['default'] = sp['name']
MdpInfo = sp['metadata']


class SP(Product):
    """ A subclass of `Product` for tests. """

    def __init__(self,
                 description='UNKNOWN',
                 typ_='SP',
                 creator='UNKNOWN',
                 version='9',
                 creationDate=FineTime(0),
                 rootCause='UNKNOWN',
                 startDate=FineTime(0),
                 endDate=FineTime(0),
                 instrument='UNKNOWN',
                 modelName='UNKNOWN',
                 mission='_AGS',
                 zInfo=None,
                 **kwds):
        metasToBeInstalled = copy.copy(locals())
        for x in ('self', '__class__', 'zInfo', 'kwds'):
            metasToBeInstalled.pop(x)

        self.zInfo = sp
        assert PPI['metadata']['version']['data_type'] == 'string'
        super().__init__(zInfo=zInfo, **metasToBeInstalled, **kwds)
        # super().installMetas(metasToBeInstalled)


def makeCal2D(width=11, height=11):

    center_x = int(width / 2)
    center_y = int(height / 2)
    z = []
    for y in range(height):
        zx = []
        for x in range(width):
            dx = x-center_x
            dy = y-center_y
            r = sqrt(dx*dx+dy*dy)
            zx.append(sin(r)/r if r else 1.0)
        z.append(zx)
    return z


class DemoProduct(MapContext):
    def __init__(self, *args, **kwds):

        super().__init__(*args, **kwds)
        self.zInfo['name'] = 'DemoProduct'
        self.zInfo['description'] = 'Test class %s.' % self.zInfo['name']
        self.zInfo['metadata']['type']['default'] = self.zInfo['name']


def get_demo_product(desc=''):
    """
    A complex context product as a reference for testing and demo.

::

|__ meta                                          <MetaData>
    |   |__ description                                 <string>
    |   |__ type                                        <string>
    |   |__ level                                       <string>
    |   |__ creator                                     <string>
    |   |__ creationDate                              <finetime>
    |   |__ rootCause                                   <string>
    |   |__ version                                     <string>
    |   |__ FORMATV                                     <string>
    |   |__ speed                                       <vector>
    |   \__ listeners                               <ListenerSet>
    |__ measurements                          <CompositeDataset>
    |   |__ meta                                      <MetaData>
    |   |   \__ listeners                           <ListenerSet>
    |   |__ Time_Energy_Pos               <TableDataset> (5, 20)
    |   |   |__ meta                                  <MetaData>
    |   |   |   |__ description                         <string>
    |   |   |   |__ shape                                <tuple>
    |   |   |   |__ type                                <string>
    |   |   |   |__ version                             <string>
    |   |   |   |__ FORMATV                             <string>
    |   |   |   \__ listeners                       <ListenerSet>
    |   |   |__ Time                              <Column> (20,)
    |   |   |__ Energy                            <Column> (20,)
    |   |   |__ Error                             <Column> (20,)
    |   |   |__ y                                 <Column> (20,)
    |   |   \__ z                                 <Column> (20,)
    |   |__ calibration                  <ArrayDataset> (11, 11)
    |   \__ dset                                           <str>
    |__ Environment Temperature              <ArrayDataset> (7,)
    |__ Browse                               <image/png> (5976,)
    |__ refs                                      <RefContainer>
    |   |__ a_reference                             <ProductRef>
    |   \__ a_different_name                        <ProductRef>
    |__ history                                        <History>
    |   |__ PARAM_HISTORY                                  <str>
    |   |__ TASK_HISTORY                                   <str>
    |   \__ meta                                      <MetaData>
    |       \__ listeners                           <ListenerSet>
    \__ listeners                                   <ListenerSet>

    """

    prodx = DemoProduct(desc if desc else 'A complex product for demo/test.')
    prodx.creator = 'Frankenstein'
    # add a parameter with validity descriptors to the product
    prodx.meta['speed'] = NumericParameter(
        description='an extra param',
        value=Vector((1.1, 2.2, 3.3)),  # do not use list
        valid={(1, 22): 'normal', (30, 33): 'fast'}, unit='meter')

    # A CompositeDataset 'measurements' of two sub-datasets: calibration and measurements
    composData = CompositeDataset()
    prodx['measurements'] = composData
    # A 2-dimensional array of calibration data
    a5 = makeCal2D()
    a8 = ArrayDataset(data=a5, unit='count', description='array in composite')
    a10 = 'calibration'

    # a tabledataset as the measurements
    ELECTRON_VOLTS = 'eV'
    SECONDS = 'sec'
    METERS = 'm'
    t = [x * 1.0 for x in range(20)]
    e = [2 * x + 100 for x in t]
    err = [random.random() * 2 - 1 for x in t]
    y = [10 * sin(x*2*3.14/len(t)) for x in t]
    z = [10 * cos(x*2*3.14/len(t)) for x in t]

    x = TableDataset(description="A table of measurement reslts")
    x["Time"] = Column(data=t, unit=SECONDS)
    x["Energy"] = Column(data=e, unit=ELECTRON_VOLTS)
    x["Error"] = Column(data=err, unit=ELECTRON_VOLTS)
    x["y"] = Column(data=y, unit=METERS)
    x["z"] = Column(data=z, unit=METERS)
    # set a tabledataset ans an arraydset, with a parameter in metadata
    # this is the first dataset in composData
    composData['Time_Energy_Pos'] = x

    # put the dataset to the compositedataset. here set() api is used
    composData.set(a10, a8)

    # an arraydsets as environment temperature
    a1 = [768, 767, 766, 4.4, 4.5, 4.6, 5.4E3]
    a2 = 'C'
    a3 = 'Environment Temperature'
    a4 = ArrayDataset(data=a1, unit=a2,
                      description='A 2D array for environment temperature')
    # metadata to the dataset
    a11 = 'T0'
    a12 = DateParameter('2020-02-02T20:20:20.0202',
                        description='meta of compositeDs')
    # This is not the best as a4.T0 does not exist
    # a4.meta[a11] = a12
    # this does it a4.T0 = a12 or:
    setattr(a4, a11, a12)
    # put the arraydataset to the product with a name a3.
    prodx[a3] = a4

    # an image as Browse
    fname = 'imageBlue.png'
    fname = op.join(op.join(op.abspath(op.dirname(__file__)),
                            'resources'), fname)
    image = loadMedia(fname)
    image.file = fname
    prodx['Browse'] = image
    return prodx


def get_related_product():
    p = Product(description='A related Product')
    p['a dumb table'] = TableDataset([('pi', 'e'), (3.14, 2.72)])
    return p



# -*- coding: utf-8 -*-

from .metadata import Parameter
from .quantifiable import Quantifiable
from .datatypes import Vector, Vector2D, Vector3D, Quaternion

from collections.abc import Sequence
from collections import OrderedDict
from copy import copy
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class NumericParameter(Parameter, Quantifiable):
    """ A Parameter that has a number as the value, a unit, and a typecode.
    """

    def __init__(self,
                 value=None,
                 description='UNKNOWN',
                 typ_='',
                 default=None,
                 unit=None,
                 valid=None,
                 typecode=None,
                 **kwds):
        """ Set up a parameter whose value is a or a list of numbers.

        typ_: type of the parameter value.
        """

        # collect args-turned-local-variables.
        args = copy(locals())
        args.pop('__class__', None)
        args.pop('kwds', None)
        args.pop('self', None)
        args.update(kwds)

        super().__init__(
            value=value, description=description, typ_=typ_, default=default, unit=unit, valid=valid, typecode=typecode)
        # Must overwrite the self._all_attrs set by supera()
        self._all_attrs = args

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           type=self._type,
                           default=self._default,
                           value=self._value,
                           valid=self._valid,
                           unit=self._unit,
                           typecode=self._typecode)

    def setValue(self, value):
        """ accept any type that a Vector does.
        """
        if value is not None and issubclass(value.__class__, Sequence):
            if self.type in ('vector', 'vector2d', 'vector3d', 'quaternion'):
                d = list(value)
                len_d = len(d)
                if len_d == 2:
                    value = Vector2D(d)
                elif len_d == 3:
                    value = Vector3D(d)
                elif len_d == 4 and getattr(self, '_type', None) == 'quaternion':
                    value = Quaternion(d)
                else:
                    value = Vector(d)
        super().setValue(value)

    def setDefault(self, default):
        """ accept any type that a Vector does.

        'Accept sequence of only 0 to 4 elements for NumericParameter and Vectors.')

        """
        if default is not None and issubclass(default.__class__, Sequence):
            # MUST NOT depend on self.type being existing
            if getattr(self, '_type', None) in ('vector', 'vector2d', 'vector3d', 'quaternion'):
                d = list(default)
                len_d = len(d)
                if len_d == 2:
                    default = Vector2D(d)
                elif len_d == 3:
                    default = Vector3D(d)
                elif len_d == 4 and getattr(self, '_type', None) == 'quaternion':
                    default = Quaternion(d)
                else:
                    pass
                    # raise ValueError(
                    #        'Sequence of only 0 to 4 elements for NumericParameter and Vectors.')
        super().setDefault(default)


class BooleanParameter(Parameter):
    """ A Parameter that has a boolean as the value.
    """

    def __init__(self,
                 value=None,
                 description='UNKNOWN',
                 default=None,
                 valid=None,
                 **kwds):
        """ Set up a parameter whose value is a boolean

        """

        typ_ = kwds.pop('typ_', 'boolean')
        # collect args-turned-local-variables.
        args = copy(locals())
        args.pop('__class__', None)
        args.pop('kwds', None)
        args.pop('self', None)
        args.update(kwds)

        super().__init__(
            value=value, description=description, typ_=typ_, default=default, valid=valid)
        # Must overwrite the self._all_attrs set by supera()
        self._all_attrs = args

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(description=self.description,
                           type=getattr(self, '_type', ''),
                           default=getattr(self, '_default', ''),
                           value=getattr(self, '_value', ''),
                           valid=getattr(self, '_valid', ''),)

    def setValue(self, value):
        """ accept any type that `bool` does.
        """
        b = None if value is None else bool(value)
        super().setValue(b)

    def setDefault(self, default):
        """ accept any type that `bool` Vector does.
        """
        b = None if default is None else bool(default)
        super().setDefault(b)



# -*- coding: utf-8 -*-

from .serializable import Serializable
from .eq import DeepEqual
from .classes import Classes
from .quantifiable import Quantifiable

from functools import lru_cache
from collections import OrderedDict
import builtins
from math import sqrt
import array
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

bltns = vars(builtins)

ENDIAN = 'little'
""" Endianess of products generated. """

# data_type in parameter/column descriptors vs `data.__class__.__name__`
DataTypes = {
    'array': 'array.array',
    'baseProduct': 'BaseProduct',
    'binary': 'int',
    'boolean': 'bool',
    'byte': 'int',
    'complex': 'complex',
    'finetime': 'FineTime',
    'finetime1': 'FineTime1',
    'float': 'float',
    'hex': 'int',
    'integer': 'int',
    'list': 'list',
    'mapContext': 'MapContext',
    'product': 'Product',
    'quaternion': 'Quaternion',
    'short': 'int',
    'string': 'str',
    'tuple': 'tuple',
    # 'numericParameter': 'NumericParameter',
    # 'dateParameter': 'DateParameter',
    # 'stringParameter': 'StringParameter',
    'vector': 'Vector',
    'vector2d': 'Vector2D',
    'vector3d': 'Vector3D',
    '': 'None'
}
""" Allowed data (Parameter and Dataset) types and the corresponding classe names.
The keys are mnemonics for humans; the values are type(x).__name__.
"""

DataTypeNames = {}
for tn, tt in DataTypes.items():
    if tt == 'int':
        DataTypeNames[tt] = 'integer'
    else:
        DataTypeNames[tt] = tn
DataTypeNames.update({
    'NoneType': '',
    'dict': 'vector',
    'OrderedDict': 'vector',
    'UserDict': 'vector',
    'ODict': 'vector',
})
""" maps class type names to human-friendly types, reverse `DataTypes`. """

del tt, tn


try:
    import numpy as np

    # numpy type to python type https://stackoverflow.com/a/34919415/13472124
    # nptype_to_pythontype = {v: getattr(bltns, k)
    #                        for k, v in np.typeDict.items() if k}
    pass
except ImportError:
    pass

# https://docs.python.org/3.6/library/ctypes.html#fundamental-data-types

ctype_to_typecode = {
    'c_bool': 't_',    # Bool
    'c_char': 'b',  # 1-char bytes
    'c_wchar': 'b',  # 1-char string
    'c_byte': 'b',  # int? signed char
    'c_ubyte': 'B',  # unsigned char
    'c_short': 'h',  # signed short
    'c_ushort': 'H',  # unsigned short
    'c_int': 'i',  # signed int
    'c_uint': 'I',  # unsigned int
    'c_long': 'l',  # signed long
    'c_ulong': 'L',  # unsigned long
    'c_longlong': 'q',  # signed long long
    'c_ulonglong': 'Q',  # unsigned long long
    'c_float': 'f',  # float
    'c_double': 'd',  # double
    'c_char_p': 'u',  # string
}

# from https://stackoverflow.com/a/53702352/13472124 by tel


def get_typecodes():
    import numpy as np
    import ctypes as ct
    import pprint
    # np.ctypeslib.as_ctypes_type(dtype)
    simple_types = [
        ct.c_byte, ct.c_short, ct.c_int, ct.c_long, ct.c_longlong,
        ct.c_ubyte, ct.c_ushort, ct.c_uint, ct.c_ulong, ct.c_ulonglong,
        ct.c_float, ct.c_double,
    ]
    return pprint.pprint([
        {np.dtype(ctype).str: ctype.__name__ for ctype in simple_types},
        {ctype_to_typecode[ctype.__name__]: np.dtype(
            ctype).itemsize for ctype in simple_types}
    ])

# generated with get_typecodes() above


# w/o endian
numpytype_to_ctype = {
    'f4': 'c_float',
    'f8': 'c_double',
    'i2': 'c_short',
    'i4': 'c_int',
    'i8': 'c_long',
    'u2': 'c_ushort',
    'u4': 'c_uint',
    'u8': 'c_ulong',
    'i1': 'c_byte',
    'u1': 'c_ubyte'
}

# generated with get_typecodes() above

typecode_itemsize = {'B': 1,
                     'H': 2,
                     'I': 4,
                     'L': 8,
                     'b': 1,
                     'd': 8,
                     'f': 4,
                     'h': 2,
                     'i': 4,
                     'l': 8}

numpy_dtypekind_to_typecode = {
    'b': 't_',  # 'boolean',
    'i': 'i',    # integer',
    'u': 'I',    # unsigned integer',
    'f': 'd',    # float',
    'c': 'c',    # complex float',
    'm': 'datetime.timedelta',    # timedelta',
    'M': 'datetime,    # datetime',
    'O': 'object',    # object',
    'S': 'B',    # string',
    'U': 'B',    # unicode string',
    'V': 'V',    # fixed chunk of memory for other type ( void )',
}


def numpytype_to_typecode(x): return ctype_to_typecode[numpytype_to_ctype[x]]


def cast(val, typ_, namespace=None):
    """Casts the input value to type specified.

    For example 'binary' type '0x9' is casted to int 9.

    Parameters
    ----------
    val : multiple
        value to be casted.
    typ_ : string
        "human name" of `datatypes.DataTypeNames`. e.g. `integer`, `vector2d`, 'list'.
    namespace : dict
        namespace to look up classes. default is `Classes.mapping`.

    Returns
    -------
    multiple
        Casted value.

    Raises
    ------

    """

    t = DataTypes[typ_]
    vstring = str(val).lower()
    tbd = bltns.get(t, None)  # lookup_bd(t)
    if tbd:
        if t == 'int':
            base = 16 if vstring.startswith(
                '0x') else 2 if vstring.startswith('0b') else 10
            return tbd(vstring, base)
        elif t == 'bytes':
            return int(val).to_bytes(1, ENDIAN)
        return tbd(val)
    else:
        return Classes.mapping[t](val) if namespace is None else namespace[t](val)


class Vector(Quantifiable, Serializable, DeepEqual):
    """ N dimensional vector.

    If description, type etc meta data is needed, use a Parameter.

    A Vector can compare with a value whose type is in ``DataTypes``, the quantity being used is the magnitude.
    """

    def __init__(self, components=None, **kwds):
        """ invoked with no argument results in a vector of
        [0, 0, 0] components.

        Parameters
        ----------

        Returns
        -------

        """
        if components is None:
            self._data = [0, 0, 0]
        else:
            self.setComponents(components)
        super().__init__(**kwds)

    @ property
    def components(self):
        """ for property getter
        Parameters
        ----------

        Returns
        -------
        """
        return self.getComponents()

    @ components.setter
    def components(self, components):
        """ for property setter

        Parameters
        ----------

        Returns
        -------

        """
        self.setComponents(components)

    def getComponents(self):
        """ Returns the actual components that is allowed for the components
        of this vector.

        Parameters
        ----------

        Returns
        -------

        """
        return self._data

    def setComponents(self, components):
        """ Replaces the current components of this vector.

        Parameters
        ----------

        Returns
        -------

        """
        # for c in components:
        #     if not isinstance(c, Number):
        #         raise TypeError('Components must all be numbers.')
        # must be list to make round-trip Json
        self._data = list(components)

    def __eq__(self, obj, verbose=False, **kwds):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return sqrt(sum(x*x for x in self._data)) == obj
        else:
            return super(Vector, self).__eq__(obj)

    def __lt__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return sqrt(sum(x*x for x in self._data)) < obj
        else:
            return super(Vector, self).__lt__(obj)

    def __gt__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return sqrt(sum(x*x for x in self._data)) > obj
        else:
            return super(Vector, self).__gt__(obj)

    def __le__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return sqrt(sum(x*x for x in self._data)) <= obj
        else:
            return super(Vector, self).__le__(obj)

    def __ge__(self, obj):
        """ can compare value """
        if type(obj).__name__ in DataTypes.values():
            return sqrt(sum(x*x for x in self._data)) >= obj
        else:
            return super(Vector, self).__ge__(obj)

    def __len__(self):
        """
        Parameters
        ----------

        Returns
        -------

        """
        return len(self._data)

    def toString(self, level=0, **kwds):
        return self.__repr__()

    __str__ = toString

    string = toString
    txt = toString

    def __getstate__(self):
        """ Can be encoded with serializableEncoder
        Parameters
        ----------

        Returns
        -------

        """
        return OrderedDict(
            components=list(self._data),
            unit=self._unit,
            typecode=self._typecode)


class Vector2D(Vector):
    """ Vector with 2-component data"""

    def __init__(self, components=None, **kwds):
        """ invoked with no argument results in a vector of
        [0, 0] components
        Parameters
        ----------

        Returns
        -------
        """
        super().__init__(**kwds)

        if components is None:
            self._data = [0, 0]
        else:
            self.setComponents(components)


class Vector3D(Vector):
    """ Vector with 3-component data"""

    def __init__(self, components=None, **kwds):
        """ invoked with no argument results in a vector of
        [0, 0, 0]] components
        Parameters
        ----------

        Returns
        -------
        """
        super().__init__(**kwds)

        if components is None:
            self._data = [0, 0, 0]
        else:
            self.setComponents(components)


class Quaternion(Vector):
    """ Quaternion with 4-component data.
    """

    def __init__(self, components=None, **kwds):
        """ invoked with no argument results in a vector of
        [0, 0, 0, 0] components
        Parameters
        ----------

        Returns
        -------
        """
        super(Quaternion, self).__init__(**kwds)

        if components is None:
            self._data = [0, 0, 0, 0]
        else:
            self.setComponents(components)



# -*- coding: utf-8 -*-

from .abstractcomposite import AbstractComposite
from .typed import Typed
from .attributable import Attributable
from .copyable import Copyable
from .serializable import Serializable
from .odict import ODict
from .dataset import Dataset
from .listener import MetaDataListener

try:
    from .unstructureddataset_datamodel import Model
except ImportError:
    Model = {'metadata': {}}
import xmltodict

import mmap

import json
from copy import copy
from collections import OrderedDict
import logging
# create logger
logger = logging.getLogger(__name__)


class UnstructuredDataset(Dataset, Copyable):
    """ Container for data without pre-defined structure or organization..

    `MetaDataListener` must stay to the left of `AbstractComposite`.

    For `xmltodict`  `xml_attribs` default to ```False```. 
    """

    def __init__(self, data=None,
                 description=None,
                 typ_=None,
                 doctype=None,
                 version=None,
                 zInfo=None,
                 alwaysMeta=True,
                 **kwds):
        """
        Accepts keyword args to `xmltodict`, e.g. `xml_attribs`, `attr_prefix` and `cdata_key`.
        """

        # collect MDPs from args-turned-local-variables.
        metasToBeInstalled = copy(locals())
        metasToBeInstalled.pop('__class__', None)
        metasToBeInstalled.pop('kwds', None)
        metasToBeInstalled.pop('self', None)
        metasToBeInstalled.pop('zInfo', None)

        self._list = []

        global Model
        if zInfo is None:
            zInfo = Model
        # for `xmltodict`  `xml_attribs` default to ```True```.
        self.xml_attribs = kwds.pop('xml_attribs', True)
        self.attr_prefix = kwds.pop('attr_prefix', '')
        self.cdata_key = kwds.pop('cdata_key', 'value')

        super().__init__(zInfo=zInfo, **metasToBeInstalled,
                         **kwds)  # initialize typ_, meta, unit
        self.data_pv = {}
        self.put(data=data, doctype=doctype)

    def getData(self):
        """ Optimized for _data being initialized to be `_data` by `DataWrapper`.

        """

        try:
            return self._data
        except AttributeError:
            return self._data.data

    def make_meta(self, print=False, **kwds):
        full = self.jsonPath('$..*', **kwds)
        for pv in full:
            self.__setattr__(p[0], p[1])
        self.data_pv = full

    def put(self, data, doctype=None, **kwds):
        """ Put data in the dataset.

        Depending on `doctype`:
        * Default is `None` for arbitrarily nested Pyton data structure.
        * Use 'json' to have the input string loaded by `json.loads`,
        * 'xml' by `xmltodict.parse`.
        """

        if doctype:
            self.doctype = doctype
        if data:
            stid = b'_STID' if issubclass(data.__class__, bytes) else '_STID'
            # do not ask for self.type unless there is real data.
            if self.doctype == 'json':
                if issubclass(data.__class__, mmap.mmap):
                    data = bytearray(data)
                loaded = json.loads(data, **kwds)
            elif self.doctype == 'xml':
                xa = kwds.pop('xml_attribs', None)
                xa = self.xml_attribs if xa is None else xa
                ap = kwds.pop('attr_prefix', None)
                ap = self.attr_prefix if ap is None else ap
                ck = kwds.pop('cdata_key', None)
                ck = self.cdata_key if ck is None else ck

                loaded = xmltodict.parse(data,
                                         attr_prefix=ap,
                                         cdata_key=ck,
                                         xml_attribs=xa, **kwds)
            else:
                # Others are loaded as data structures
                loaded = data
                #raise ValueError('Cannot process %s type.' % str(doctype))
            # set Escape if not set already
            if hasattr(loaded, '__iter__') and stid in loaded:
                ds = loaded[stid]
                if not ds.startswith('0'):
                    loaded[stid] = '0%s' % ds
        else:
            loaded = data
        super().setData(loaded)
        # self.make_meta()

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(
            meta=getattr(self, '_meta', None),
            data=self.getData())



# -*- coding: utf-8 -*-

from ..utils.common import findShape

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class Shaped():
    """ An object of specifiable kinds.

    This class is for parameters, not for products/datasets to make 'shape' MDP.
    """

    def __init__(self, **kwds):
        """ Has a shape.
        Parameters
        ----------

        Returns
        -------
        """

        # print(__name__ + str(kwds))
        super().__init__(**kwds)

    def updateShape(self, **kwds):

        self.__setattr__('shape', findShape(self.data, **kwds))
        # this is not working for the firist call as shape is not in
        # `attributable.__init__`s initializing list because 'shape'
        # is not in `ArrayDAtaset.init` args.
        # self.shape = findShape(self.data, **kwds)

        return self.shape



# -*- coding: utf-8 -*-

from ..utils.common import trbk
from ..utils.moduleloader import SelectiveMetaFinder, installSelectiveMetaFinder
from .namespace import Load_Failed, NameSpace_meta

import builtins
from collections import ChainMap
import sys
import logging
import copy
import importlib
from functools import lru_cache

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


''' Note: this has to be in a different file where other interface
classes are defined to avoid circular dependency (such as ,
Serializable.
'''

# modules and classes to import from them

Modules_Classes = {
    'fdi.dataset.deserialize': ['deserialize'],
    'fdi.dataset.listener': ['ListenerSet'],
    'fdi.dataset.serializable': ['Serializable'],
    'fdi.dataset.eq': ['DeepEqual'],
    'fdi.dataset.odict': ['ODict'],
    'fdi.dataset.finetime': ['FineTime', 'FineTime1', 'utcobj'],
    'fdi.dataset.history': ['History'],
    'fdi.dataset.baseproduct': ['BaseProduct'],
    'fdi.dataset.product': ['Product'],
    'fdi.dataset.browseproduct': ['BrowseProduct'],
    'fdi.dataset.testproducts': ['TP', 'TP_0X', 'TB', 'TC', 'TCC', 'TM', 'SP', 'DemoProduct'],
    'fdi.dataset.datatypes': ['Vector', 'Vector2D', 'Vector3D', 'Quaternion'],
    'fdi.dataset.metadata': ['AbstractParameter', 'Parameter', 'MetaData'],
    'fdi.dataset.numericparameter': ['NumericParameter', 'BooleanParameter'],
    'fdi.dataset.dateparameter': ['DateParameter'],
    'fdi.dataset.stringparameter': ['StringParameter'],
    'fdi.dataset.arraydataset': ['ArrayDataset', 'Column'],
    'fdi.dataset.mediawrapper': ['MediaWrapper'],
    'fdi.dataset.dataset': ['GenericDataset', 'CompositeDataset'],
    'fdi.dataset.tabledataset': ['TableDataset', 'IndexedTableDataset'],
    'fdi.dataset.unstructureddataset': ['UnstructuredDataset'],
    'fdi.dataset.readonlydict': ['ReadOnlyDict'],
    'fdi.pal.context': ['AbstractContext', 'Context',
                        'MapContext',
                        'RefContainer',
                        'ContextRuleException'],
    'fdi.pal.urn': ['Urn'],
    'fdi.pal.productref': ['ProductRef'],
    'fdi.pal.query':  ['AbstractQuery', 'MetaQuery', 'StorageQuery'],
    # 'fdi.utils.common': ['UserOrGroupNotFoundError'],
}

Class_Module_Map = dict((c, m)
                        for m, clses in Modules_Classes.items() for c in clses)


def importModuleClasses(scope=None, mapping=None,
                        exclude=None, ignore_error=False,
                        verbose=False):
    """ Return of a set of deserializable classes in `initial` map, which is
    maintained by hand.

    Do nothing if the classes mapping is already made so repeated
    calls will not cost  more time.

    Parameters
    ----------

    scope : NoneType, string, list
        if is a string, take as a class name; (not implemented: If is None, import all classes in `initial` map; if a list, a list of class names.)
    mapping: mapping
        A mapping to get module names with class names. For a given key, if `mapping` returns a class object, the object will be the value returned for the key; if `mapping` returns a string, take it as a fully qualified module name and load its member classes; if is a list, take it as a list of module names.
    exclude : list
        A list of class names (without '.') that are not to be imported. Default is empty.
    ignore_error : boolean
        Importing errors will be logged but otherwise ignored. Default is `False`.

    Returns
    -------
    dict:
       Key and load-result pairs. load-result is `.namespace.Load_Failed` if loading of the key was not successful.
    """

    if exclude is None:
        exclude = []

    if scope is None:
        return {}
    # elif issubclass(scope.__class__, str):
    #    scope = [scope]
    if mapping is None:
        mapping = Class_Module_Map

    SelectiveMetaFinder.exclude = exclude
    if exclude or SelectiveMetaFinder.exclude:
        msg = 'With %s excluded.. and SelectiveMetaFinder.exclude=%s' % (
            str(exclude), str(SelectiveMetaFinder.exclude))
        logger.debug(msg)

    res = {}
    for cl in [scope]:
        if cl not in mapping:
            res[cl] = Load_Failed
            continue
        module_name = mapping[cl]
        if isinstance(module_name, type):
            # no need to load
            res[cl] = module_name
            continue
        # if we cannot find the module we make a class list
        class_list = Modules_Classes.get(module_name, [cl])
        left = [x for x in class_list if x not in exclude]
        if len(left) == 0:
            continue
        msg = 'importing %s from %s...' % (str(class_list), module_name)

        try:
            # m = importlib.__import__(module_name, globals(), locals(), class_list)
            m = importlib.import_module(module_name)
        except SelectiveMetaFinder.ExcludedModule as e:
            msg += ' Did not import %s, as %s' % (str(class_list), str(e))
            # ety, enm, tb = sys.exc_info()
        except SyntaxError as e:
            msg += ' Could not import %s, as %s' % (
                str(class_list), str(e))
            logger.error(msg)
            raise
        except ModuleNotFoundError as e:
            msg += ' Could not import %s, as %s' % (
                str(class_list), str(e))
            msg += '*** %s ' % str(sys.path)

            if ignore_error:
                msg += ' Ignored.'
            else:
                logger.error(msg)
                raise
        else:
            for n in left:
                res[n] = getattr(m, n)

        if verbose:
            logger.info(msg)
        else:
            logger.debug(msg)

    return res


def load(key, mapping, remove=True,
         exclude=None, ignore_error=False,
         verbose=False):

    res = importModuleClasses(key, mapping=mapping,
                              exclude=exclude,
                              ignore_error=ignore_error,
                              verbose=verbose
                              )
    return res


# add builtins (without any that starts with a '_'.
_bltn = dict((k, v) for k, v in vars(builtins).items() if k[0] != '_')

All_Globals_Builtins = ChainMap(copy.copy(globals()), _bltn)
"""A name-class `dict` of all global and builtins."""

All_Exceptions = dict((
    (n, e) for n, e in All_Globals_Builtins.items() if issubclass(e.__class__, type) and issubclass(e, Exception)))
"""A name-class `dict` of all Exceptions."""


class Classes(metaclass=NameSpace_meta,
              sources=[Class_Module_Map],
              extensions=[All_Globals_Builtins],
              load=load
              ):
    """ A dictionary of class names and their class objects that are allowed to be deserialized.

    An fdi package built-in dictionary is loaded from the module `Class_Module_Map`. Users who need add more deserializable class can for example:

    """
    pass


Class_Look_Up = Classes.mapping



# -*- coding: utf-8 -*-

from fdi.dataset.yaml2python import read_yaml
import os.path as op
import datetime
import fractions
import decimal
import traceback
from pprint import pprint
import copy
import json
import sys
import threading
import functools
import time
import locale
import array
from math import sqrt
from datetime import timezone
import pytest

from fdi.dataset.annotatable import Annotatable
from fdi.dataset.copyable import Copyable
from fdi.dataset.odict import ODict
from fdi.dataset.eq import deepcmp
from fdi.dataset.classes import Classes
from fdi.dataset.serializable import serialize
from fdi.dataset.deserialize import deserialize
from fdi.dataset.quantifiable import Quantifiable
from fdi.dataset.listener import EventSender, EventTypes, EventType, EventTypeOf, MetaDataListener, EventListener
from fdi.dataset.messagequeue import MqttRelayListener, MqttRelaySender
from fdi.dataset.composite import Composite
from fdi.dataset.metadata import Parameter, MetaData, make_jsonable
from fdi.dataset.metadataholder import MetaDataHolder
from fdi.dataset.numericparameter import NumericParameter, BooleanParameter
from fdi.dataset.stringparameter import StringParameter
from fdi.dataset.dateparameter import DateParameter
from fdi.dataset.datatypes import DataTypes, DataTypeNames
from fdi.dataset.attributable import Attributable
from fdi.dataset.abstractcomposite import AbstractComposite
from fdi.dataset.datawrapper import DataWrapper, DataWrapperMapper
from fdi.dataset.arraydataset import ArrayDataset, Column
from fdi.dataset.mediawrapper import MediaWrapper
from fdi.dataset.tabledataset import TableDataset
from fdi.dataset.dataset import Dataset, CompositeDataset
from fdi.dataset.indexed import Indexed
from fdi.dataset.ndprint import ndprint
from fdi.dataset.datatypes import Vector, Vector2D, Quaternion
from fdi.dataset.invalid import INVALID
from fdi.dataset.finetime import FineTime, FineTime1
from fdi.dataset.history import History
from fdi.dataset.baseproduct import BaseProduct
from fdi.dataset.product import Product
from fdi.dataset.browseproduct import BrowseProduct
from fdi.dataset.readonlydict import ReadOnlyDict
from fdi.dataset.unstructureddataset import UnstructuredDataset
from fdi.dataset.testproducts import SP, get_demo_product
from fdi.utils.checkjson import checkjson
from fdi.utils.loadfiles import loadMedia
from fdi.utils.ydump import ydump

from jsonpath_ng.parser import JsonPathParserError


if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False

locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

# make format output in /tmp/outputs.py
mk_outputs = 0
output_write = 'tests/outputs.py'

if mk_outputs:
    with open(output_write, 'wt', encoding='utf-8') as f:
        f.write('# -*- coding: utf-8 -*-\n')

if __name__ == '__main__' and __package__ is None:
    # run by python3 tests/test_dataset.py

    if not mk_outputs:
        from outputs import nds2, nds3, out_Dataset, out_ArrayDataset, out_TableDataset, out_CompositeDataset, out_FineTime, out_MetaData
else:
    # run by pytest

    # This is to be able to test w/ or w/o installing the package
    # https://docs.python-guide.org/writing/structure/
    #from pycontext import fdi
    import fdi
    if not mk_outputs:
        from outputs import nds2, nds3, out_Dataset, out_ArrayDataset, out_TableDataset, out_CompositeDataset, out_FineTime, out_MetaData

    import logging
    import logging.config
    # create logger
    if 1:
        from logdict import logdict
        logging.config.dictConfig(logdict)
    else:
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(name)8s %(process)d %(threadName)s %(levelname)s %(funcName)10s() %(lineno)3d- %(message)s')

    logger = logging.getLogger()
    logger.debug('logging level %d' % (logger.getEffectiveLevel()))
    # logging.getLogger().setLevel(logging.DEBUG)

    logging.getLogger("requests").setLevel(logging.WARN)
    logging.getLogger("urllib3").setLevel(logging.WARN)
    logging.getLogger("filelock").setLevel(logging.WARN)


def test_read_yaml():
    resrc_dir = op.join(op.abspath(op.dirname(__file__)), 'resources')
    descriptor, input_files = read_yaml(resrc_dir, verbose=False)

    pprint(descriptor)



# -*- coding: utf-8 -*-

""" from https://stackoverflow.com/a/70797664  reubano"""

from .namespace import NameSpace_meta, refloader
from ..utils.common import find_all_files, trbk

#from jsonschema import Draft7Validator as the_validator
from jsonschema import Draft201909Validator as the_validator
from jsonschema import RefResolver
from jsonschema.exceptions import RefResolutionError

import json
import os
import copy
from collections import ChainMap
import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

FDI_SCHEMA_BASE = 'https://fdi.schemas'

FDI_SCHEMA_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../schemas'))
""" The directory where the schema definition files are stored."""

""" The id-obj map for package-wide schemas. To be updated by `makeSchemaStore` when it runs for the first time."""


def makeSchemaStore(schema_dir=None, verbose=False):
    """make a mapping of schema id and schema obj loaded from the given directory.

    Parameters
    ----------
    schema_dir : str
        Name of a directory containing schema definitions in JSON files
    and subdirectories. If is `None` set to `FDI_SCHEMA_DIR`.
    verbose : bool
        Say more if set.

    Returns
    -------
    dict
        file paths vs. schema objects

    Raises
    ------
    # raise ValueError

    """

    if schema_dir is None:
        # make package schemas list
        schema_dir = FDI_SCHEMA_DIR

    dirs = find_all_files(schema_dir, verbose=verbose,
                          include='**/*.js*n',
                          exclude=(''),
                          absdir=True)

    schemas = []
    for source in dirs:
        try:
            with open(source, 'r') as f:
                schemas.append(json.load(f))
        except json.decoder.JSONDecodeError as e:
            logger.warning('Cannot load schema %s. No Skipping...' % source)
            logger.warning(trbk(e))
            raise
    store = dict((schema.get("$id",schema.get("id")), schema) for schema in schemas)

    return store


def schema_dir_loader(key, mapping, remove=True,
                      exclude=None, ignore_error=False,
                      verbose=False):
    """ load all schemas in the given directory and its subdirs. """

    if exclude is None:
        exclude = []

    if key is None:
        return {}

    res = {}

    spath, sname = tuple(key.rsplit('/', 1))
    family = (x for x in mapping if x not in exclude and x.startswith(spath))
    for sch in family:
        jsn = json.load(sch)
        if jsn is None:
            if ignore_error:
                continue
            else:
                raise ValueError(sch)
        else:
            res[sch] = jsn
            del mapping[sch]
    return res


class Schemas(metaclass=NameSpace_meta,
              sources=[makeSchemaStore(FDI_SCHEMA_DIR)],
              load=refloader
              ):
    pass


def getValidator(schema, schemas=None, schema_store=None, base_schema=None, verbose=False):
    """ Returns a `jsonschema` validator that knows where to find given schemas.

    :schema: the schema this validator is made for.
    :schemas: A map of schema id and schema objects. default is all schemas found in ```schema_store```. If it has '$schema' and '$id' as keys, it will be the lone schema to be validated against by the returned validator.
    :schema_store: get schemas here if ```schemas``` is ```None```. default is `FDI_SCHEMA_STORE`.
    :base_schema: A reference schema object providing BaseURI. Default is `schema_store[FDI_SCHEMA_BASE]`, whereas the `schema_store` is made with `FDI_SCHEMA_STORE` and `schemas`.
    """

    if issubclass(schema.__class__, str):
        schema = json.loads(schema)
    the_validator.check_schema(schema)
    if issubclass(schemas.__class__, dict) and '$schema' in schemas and '$id' in schemas:
        store = {schemas['$id']: schemas}
    else:
        if schemas is None:
            schemas = {}
        if schema_store is None:
            schema_store = Schemas.mapping
        store = schema_store.add_ns(schemas, 0) if schemas else schema_store
    if verbose:
        print('Schema store:', list(store))
    if base_schema is None:
        # json.load(open("schema/dir/extend.schema.json"))
        #resolver = RefResolver(schema['$id'], store=store, referrer=schema)
        if FDI_SCHEMA_BASE not in store:
            raise ValueError(
                'Base schema is not given and FDI_SCHEMA_BASE %s is not in store.' % FDI_SCHEMA_BASE)
        base_schema = store[FDI_SCHEMA_BASE]
    resolver = RefResolver.from_schema(base_schema, store=store)

    if verbose:
        print('Schema resolver:', resolver)
    validator = the_validator(schema, resolver=resolver)

    return validator


def validateJson(data, validator):
    """ validates a JSON object.

    :data: a JSON object or a _file_full_path that ends with 'json' or 'jsn'.
    """

    if issubclass(data.__class__, str) and \
       (data.endswith('.jsn') or data.endswith('.json')):
        instance = json.load(open(data))
    else:
        instance = data
    try:
        errors = sorted(validator.iter_errors(instance), key=lambda e: e.path)
    except RefResolutionError as e:
        logger.error(e)
    return errors


class Schmas(ChainMap):
    def __init__(self, **kwds):

        makeSchemaStore()



# -*- coding: utf-8 -*-

from ..dataset.baseproduct import BaseProduct
from ..dataset.classes import Classes
from ..dataset.product import Product
from ..dataset.serializable import serialize
from ..dataset.deserialize import deserialize, Class_Look_Up
from .urn import Urn, parseUrn, parse_poolurl, makeUrn
from .versionable import Versionable
from .taggable import Taggable
from . import dicthk
from .definable import Definable
from ..utils.common import (fullname, lls, trbk, pathjoin,
                            logging_ERROR,
                            logging_WARNING,
                            logging_INFO,
                            logging_DEBUG
                            )
from .productref import ProductRef
from .query import AbstractQuery, MetaQuery, StorageQuery

from collections import OrderedDict, ChainMap
from functools import lru_cache
import logging
import filelock
import getpass
import functools
import os
import sys
import builtins

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    from urllib.parse import urlparse
else:
    PY3 = False
    from urlparse import urlparse

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

Lock_Path_Base = '/tmp/fdi_locks_' + getpass.getuser()
# lock time-out
locktout = 10


def makeLockpath(direc, op='w'):
    """ returns the appropriate path to put lock file.

    creats the path if non-existing. Set lockpath-base permission to all-modify so other fdi users can use.
    op: 'r' for readlock no-reading) 'w' for writelock (no-writing)
    """
    if not os.path.exists(Lock_Path_Base):
        os.makedirs(Lock_Path_Base, mode=0o777)

    lp = pathjoin(Lock_Path_Base, direc.replace('/', '_'))

    return lp+'.read' if op == 'r' else lp+'.write'


class ProductPool(Definable, Taggable, Versionable):
    """ A mechanism that can store and retrieve Products.

A product pool should not be used directly by users. The general user should access data in a ProductPool through a ProductStorage instance.

When implementing a ProductPool, the following rules need to be applied:

    1. Pools must guarantee that a Product saved via the pool saveProduct(Product) method is stored persistently, and that method returns a unique identifier (URN). If it is not possible to save a Product, an IOException shall be raised.
    2. A saved Product can be retrieved using the loadProduct(Urn) method, using as the argument the same URN that assigned to that Product in the earlier saveProduct(Product) call. No other Product shall be retrievable by that same URN. If this is not possible, an IOException or GeneralSecurityException is raised.
    3. Pools should not implement functionality currently implemented in the core package. Specifically, it should not address functionality provided in the Context abstract class, and it should not implement versioning/cloning support.

    """

    def __init__(self, poolname='', poolurl='', **kwds):
        """
        Creates and initializes a productpool.

        * poolname: if provided will override that in poolurl.
        * poolurl: needed to initialize.

        """
        super().__init__(**kwds)

        self.setPoolname(poolname)
        self.setPoolurl(poolurl)
        # self._pathurl = pr.netloc + pr.path
        # self._pathurl = None
        self._poolmanager = None
        self.ignore_error_when_delete = False

    class ParametersIncommpleteError(Exception):
        pass

    def setup(self):
        """ Sets up interal machiney of this Pool,
        but only if self._poolname and self._poolurl are present,
        and other pre-requisits are met.

        Subclasses should implement own setup(), and
        make sure that self._poolname and self._poolurl are present with ``

        if <pre-requisit not met>: return True
        if super().setup(): return True

        # super().setup() has done its things by now.
        <do setup>
        return False
``
        returns: True if not both  self._poolname and self._poolurl are present.

        """

        if not hasattr(self, '_poolurl') or not self._poolurl:
            return True

        return False

    @property
    def poolname(self):
        """ for property getter
        """
        return self.getPoolname()

    @poolname.setter
    def poolname(self, poolname):
        """ for property setter
        """
        self.setPoolname(poolname)

    def getPoolname(self):
        """ Gets the poolname of this pool as an Object. """
        return self._poolname

    def setPoolname(self, poolname):
        """ Replaces the current poolname of this pool.
        """
        self._poolname = poolname

    @property
    def poolurl(self):
        """ for property getter
        """
        return self.getPoolurl()

    @poolurl.setter
    def poolurl(self, poolurl):
        """ for property setter
        """
        self.setPoolurl(poolurl)

    def getPoolurl(self):
        """ Gets the poolurl of this pool as an Object. """
        return self._poolurl

    def setPoolurl(self, poolurl):
        """ Replaces the current poolurl of this pool.
        """
        s = (not hasattr(self, '_poolurl') or not self._poolurl)
        self._poolpath, self._scheme, self._place, \
            self._poolname, self._username, self._password = \
            parse_poolurl(poolurl)
        self._poolurl = poolurl
        # call setup only if poolurl was None
        if s:
            self.setup()

    def accept(self, visitor):
        """ Hook for adding functionality to object
        through visitor pattern."""
        visitor.visit(self)

    def getPoolManager(self):
        """
        """
        return self._poolmanager

    def setPoolManager(self, pm):
        """
        """
        self._poolmanager = pm

    def dereference(self, ref):
        """
        Decrement the reference count of a ProductRef.

        XXX TODO
        """

        raise (NotImplementedError)

    def exists(self, urn):
        """
        Determines the existence of a product with specified URN.
        """

        raise (NotImplementedError)

    def getDefinition(self):
        """
        Returns pool definition info which contains pool type and other pool specific configuration parameters
        """
        return super().getDefinition()

    def getId(self):
        """
        Gets the identifier of this pool.
        """
        return self._poolname

    def getPoolurl(self):
        """
        Gets the pool URL of this pool.
        """
        return self._poolurl

    def getPlace(self):
        """
        Gets the place of this pool.
        """
        return self._place

    def getProductClasses(self):
        """
        Returns all Product classes found in this pool.
        mh: returns an iterator.
        """
        raise (NotImplementedError)

    def getReferenceCount(self, ref):
        """
        Returns the reference count of a ProductRef.
        """
        raise (NotImplementedError)

    def getScheme(self):
        """
        Gets the scheme of this pool.
        """
        return self._scheme

    def getUrnId(self):
        """
        Get the identifier of this pool used to build URN, usually it's same as id returned by getId().
        """
        return self.getId()

    @staticmethod
    def vectorize(*p):
        """
      ::
        vectorize(9, [8,7,6]) -> ([9, 9, 9], [8, 7,  6], True)
        """

        lens = [len(v) if isinstance(v, (list, tuple)) else 0 for v in p]

        sz = max(lens)
        # remove redundant
        s = set(lens)
        alist = any(s)
        # remove longest and scalar
        s.remove(sz)
        try:
            s.remove(0)
        except KeyError:
            pass
        if len(s):
            # found more than 2 sizes
            raise ValueError(f'Some args have different sizes {s}.')
        if sz == 0:
            # force scalar  to vector
            sz = 1

        res = [q if l else ([q] * sz) for q, l in zip(p, lens)]
        res.append(alist)
        return tuple(res)

    def isAlive(self):
        """
        Test if the pool is capable of responding to commands.
        """
        return True

    def isEmpty(self):
        """
        Determines if the pool is empty.
        """

        raise NotImplementedError

    def schematicSave(self, products, tag=None, geturnobjs=False, serialize_in=True, serialize_out=False, asyn=False, **kwds):
        """ to be implemented by subclasses to do the scheme-specific saving
        """
        raise (NotImplementedError)

    def saveProduct(self, product, tag=None, geturnobjs=False, serialize_in=True, serialize_out=False, asyn=False, **kwds):
        """
        Saves specified product(s) and returns the designated ProductRefs or URNs.

        Saves a product or a list of products to the pool, possibly under the
        supplied tag(s), and returns the reference (or a list of references if
        the input is a list of products), or Urns if geturnobjs is True.

        See pal document for pool structure.

        Parameters
        ----------
        product : BaseProduct, list
            Product or a list of them or '[ size1, prd, size2, prd2, ...]'.
        tag : str, list
            If given a tag, all products will be having this tag.
        If a list tags are given to every one product then the
        number of tags must not be the same to that of `product`. If
        they are equal, each tag is goven to the product at the same
        index in the `product` list.
        serialize_out : bool
            if `True` returns contents in serialized form.
        serialize_in : bool
            If set, product input is serialized.

        """

        res = self.schematicSave(product, tag=tag,
                                 geturnobjs=geturnobjs,
                                 serialize_in=serialize_in,
                                 serialize_out=serialize_out,
                                 asyn=asyn, **kwds)
        if issubclass(product.__class__, str) or isinstance(product, list) and \
           issubclass(product[0].__class__, str):
            # p is urn string from server-side LocalPool
            return res

        if isinstance(res, list):
            for p, u in zip(product, res):
                p._urn = u if geturnobjs else u.getUrnObj()
        else:
            product._urn = res if geturnobjs else res.getUrnObj()
        return res

    def loadDescriptors(self, urn):
        """
        Loads the descriptors belonging to specified URN.
        """

        raise (NotImplementedError)

    def schematicLoad(self, resourcetype, index,
                      start=None, end=None, serialize_out=False):
        """ to be implemented by subclasses to do the scheme-specific loading
        """
        raise (NotImplementedError)

    def loadProduct(self, urn, serialize_out=False, asyn=False):
        """
        Loads a Product belonging to specified URN.

        serialize_out: if True returns contents in serialized form.
        """
        poolname, resource, index = parseUrn(urn)
        if poolname != self._poolname:
            raise (ValueError('wrong pool: ' + poolname +
                              ' . This is ' + self._poolname))
        ret = self.schematicLoad(
            resourcetype=resource, index=index, serialize_out=serialize_out)

        if issubclass(ret.__class__, str) or isinstance(ret, list) and \
           issubclass(ret[0].__class__, str):
            # ret is a urn string from server-side LocalPool
            return ret

        if isinstance(ret, list):
            logger.warning('TODO: unexpected')
            for x, u in zip(ret, urn):
                x._urn = u
        else:
            ret._urn = urn

        return ret

    def meta(self, urn):
        """
        Loads the meta-data belonging to the product of specified URN.
        """

        raise (NotImplementedError)

    @property
    def count(self):
        """ for property getter
        """
        return self.getCount()

    @count.setter
    def count(self, count):
        """ for property setter
        """
        raise ValueError('Pool.count is read-only.')

    def getCount(self, typename=None):
        """
        Return the number of URNs for the product type.
        """

        raise (NotImplementedError)

    def reference(self, ref):
        """
        Increment the reference count of a ProductRef.
        """

        raise (NotImplementedError)

    def schematicRemove(self, urn=None, resourcetype=None, index=None, asyn=False, **kwds):
        """ to be implemented by subclasses to do the scheme-specific removing
        """
        raise (NotImplementedError)

    def remove(self, urn=None, resourcetype=None, index=None, ignore_error=False, asyn=False, **kwds):
        """
        Removes a Product belonging to specified URN or a pair of data type and serial number.
        """
        self.ignore_error_when_delete = ignore_error
        res = self.schematicRemove(
            urn, resourcetype=resourcetype, index=index, asyn=asyn, **kwds)
        return res

    def schematicWipe(self):
        """ to be implemented by subclasses to do the scheme-specific wiping.
        """
        raise (NotImplementedError)

    def removeAll(self, ignore_error=False, asyn=False, **kwds):
        """
        Remove all pool data (self, products) and all pool meta data (self, descriptors, indices, etc.).
        """
        self.ignore_error_when_delete = ignore_error
        return self.schematicWipe(asyn=asyn, **kwds)

    def saveDescriptors(self, urn, desc):
        """
        Save/Update descriptors in pool.
        """
        raise (NotImplementedError)

    def schematicSelect(self,  query, previous=None):
        """
        to be implemented by subclasses to do the scheme-specific querying.
        """
        raise (NotImplementedError)

    def select(self,  query, variable='m', ptype=Product,
               previous=None):
        """Returns a list of references to products that match the specified query.

        Parameters
        ----------
        query : str
            the 'where' query string to make a query object.
        variable : str
            name of the dummy variable in the query string.
            if `variable` is 'm', query goes via `MetaQuery(ptype, query)` ; else by `AbstractQuery(ptype, variable, query)` .
        ptype : class
            The class object whose instances are to be queried. Or
            fragment of the name of such classes.
        previous : list or str
            of urns, possibly from previous search. or a string of comma-separated urns, e.g. `'urn:a:foo:12,urn:b:bar:9'`

        Returns
        -------
        list
            of found URNs.

        """
        if issubclass(previous.__class__, str):
            previous = previous.split(',')
        if issubclass(query.__class__, StorageQuery):
            res = self.schematicSelect(query, previous)
            return res
        if issubclass(ptype.__class__, str):
            for cn, cls in Classes.mapping.items():
                if ptype in cn and issubclass(cls, BaseProduct):
                    break
            else:
                raise (ValueError(ptype + ' is not a product type.'))
            ptype = cls
        if variable == 'm':
            res = self.schematicSelect(MetaQuery(ptype, where=query), previous)
        else:
            res = self.schematicSelect(AbstractQuery(
                ptype, where=query, variable=variable), previous)
        return res

    def qm(self, qw, prod='BaseProduct', urns=None):
        """ short-hand method for `select(qw, variable'm', ptype=prod, previous=urns`.

        example:
        ..code:
        curl http://foo.edu:23456/data/pool/api/qm__m["age"]>66 and m["name"]=="Bob"'
        """
        return self.select(qw, variable='m', ptype=prod, previous=urns)

    def __repr__(self):
        co = ', '.join(str(k) + '=' + lls(v, 40)
                       for k, v in self.__getstate__().items())
        # co = ', '.join(str(k)+'=' + (v if issubclass(v.__class__, str) else
        #                              '<' + v.__class__.__name__+'>') \
        #                for k, v in self.__getstate__().items())
        return '<'+self.__class__.__name__ + ' ' + co + '>'

    def __getstate__(self):
        """ returns an odict that has all state info of this object.
        Subclasses should override this function.
        """
        return OrderedDict(
            poolurl=self._poolurl if hasattr(self, '_poolurl') else None,
        )

###########################


class PoolNotFoundError(Exception):
    pass


def _eval(code='', m='', **kwds):
    """ evaluate compiled code with given local vars. """
    try:
        res = eval(code)
    except NameError:
        res = False
        logger.debug("Evaluation error: %s. Traceback %s" %
                     (str(e), trbk(e)))

    return res


def populate_pool2(tags, ptype, sn=None, dTypes=None, dTags=None):
    """
    tags : list
        The tags in a list. `None` and empty tags are ignored.
    ptype : str
        The product name / datatype / class name of the data item, new or existing.
    sn : str
        Serial number. If is `None`, use the one in `dTypes`.
    Returns
    -------
    tuple
        dTypes and dTags with updates, and the index/serial number
    """

    # new ###
    if dTypes is None:
        dTypes = {}
    if dTags is None:
        dTags = {}

    if ptype in dTypes:
        if sn is None:
            int_sn = dTypes[ptype]['currentSN'] + 1
        else:
            int_sn = int(sn)
    else:
        int_sn = 0 if sn is None else int(sn)
        dTypes[ptype] = {
            'currentSN': int_sn,
            'sn': {}
        }

    snd = dTypes[ptype]['sn']
    _sn = str(int_sn)
    if int_sn not in snd:
        snd[int_sn] = {
            'tags': [],
            'meta': []
        }

    dTypes[ptype]['currentSN'] = int_sn

    # /new #####
    if tags is not None:
        for t in tags:
            dicthk.add_tag_datatype_sn(t, ptype, int_sn, dTypes, dTags)

    return dTypes, dTags, _sn


# Do not include leading or trailing whitespace as they are not guarantteed.
MetaData_Json_Start = '{"_ATTR_meta":'
MetaData_Json_End = '"_STID": "MetaData"}'


class ManagedPool(ProductPool, dicthk.DictHk):
    """ A ProductPool that manages its internal house keeping. """

    def __init__(self, **kwds):
        super().__init__(**kwds)
        # {type|classname -> {'sn:[sn]'}}

    def setup(self):
        """ Sets up interal machiney of this Pool,
        but only if self._poolname and self._poolurl are present,
        and other pre-requisits are met.

        Subclasses should implement own setup(), and
        make sure that self._poolname and self._poolurl are present with ``

        if <pre-requisit not met>: return True
        if super().setup(): return True

        # super().setup() has done its things by now.
        <do setup>
        return False
``
        returns: True if not both  self._poolname and self._poolurl are present.

        """

        if super().setup():
            return True
        self._classes = dict()
        # new ##
        self._dTypes = dict()
        return False

    def getPoolpath(self):
        """
        Gets the poolpath of this pool.

        poolpath is usually derived from poolurl received from ``PoolManager`` during initialization.
        """
        return self._poolpath

    def lockpath(self, op='w'):
        """ Make lock path using transformed poolname as name.

        """
        return makeLockpath(self.transformpath(self._poolname), op)

    @ lru_cache(maxsize=32)
    def transformpath(self, path):
        """ override this to changes the output from the input one (default) to something else.

        """
        if path is None:
            return None
        base = self._poolpath
        if base != '':
            if path[0] == '/':
                path = base + path
            else:
                path = base + '/' + path
        return path

    def getCacheInfo(self):
        info = {}
        for i in ['transformpath']:
            info[i] = getattr(self, i).cache_info()

        return info

    def dereference(self, ref):
        """
        Decrement the reference count of a ProductRef.
        """
        # new ###
        poolname, dt, sn = parseUrn(urn, int_index=True)
        # assert self._urns[ref.urn]['refcnt'] == self._dType[dt]['sn'][sn]['refcnt']
        r = self._dType[dt]['sn'][sn]
        if r['refcnt'] == 0:
            raise ValueError('Cannot deref below 0.')
        else:
            r['refcnt'] -= 1
        # /new ###
        # self._urns[ref.urn]['refcnt'] -= 1

    def exists(self, urn, resourcetype=None, index=None):
        """
        Determines the existence of a product with specified URN.
        """
        # new ###
        try:
            urn, datatype, sn = self.get_missing(
                urn, resourcetype, index, no_check=True)
        except KeyError:
            return False
        # return True
        # /new#
        return urn in self._urns

    def getProductClasses(self):
        """
        Returns all Product classes found in this pool.
        mh: returns an iterator.
        """
        # new ###
        assert list(self._classes.keys()) == list(self._dTypes.keys())

        return self._classes.keys()

    def getCount(self, typename=None):
        """
        Return the number of URNs for the product type.
        """
        try:
            # new ###
            # assert len(self._classes[typename]['sn']) == len(
            # self._dTypes[typename]['sn'])
            if typename:
                return len(self._dTypes[typename]['sn'])
            else:
                return sum(len(dt['sn']) for dt in self._dTypes.values())
        except KeyError:
            return 0

    def doSave(self, resourcetype, index, data, tags=None, serialize_in=True, **kwds):
        """ to be implemented by subclasses to do the action of saving
        """
        raise (NotImplementedError)

    def getReferenceCount(self, ref):
        """
        Returns the reference count of a ProductRef.
        """
        # new ###
        poolname, dt, sn = parseUrn(urn, int_index=False)
        _snd = self._dType[dt]['sn'][sn]
        # assert self._urns[ref.urn]['refcnt'] == _snd['refcnt']
        return _snd['refcnt']
        # /new ###
        # return self._urns[ref.urn]['refcnt']

    def isEmpty(self):
        """
        Determines if the pool is empty.
        """
        # new ###
        # assert len(self._urns) == len(self._dTypes)
        # return len(self._urns) == 0
        return len(self._dTypes) == 0

    def loadDescriptors(self, urn, resourcetype=None, index=None):
        """
        Loads the descriptors belonging to specified URN.
        """
        # new ###
        urn, datatype, sn = self.get_missing(urn, resourcetype, index)
        return self._dTypes
        # return self._urns[urn]

    def readHK():
        """ Subclass should overide this.

        Returns
        -------
        tuple
          Of 5 dicts that are the legacy `self._classes`, `self._tags`,
           `self._urns`, and
                    `self._dTypes`, `self._dTags`

        """
        raise NotImplementedError()

    def setMetaByUrn(self, start, end, urn=None, datatype=None, sn=None):
        """
        Sets the location of the meta data of the specified data to the given URN or a pair of data type and serial number.

        :data: usually un/serialized Product.

        Return
        :urn: stirng
        :datatype: class name
        :sn: serial number in string.
        """

        raise NotImplementedError()

    def getMetaByUrn(self, urn=None, resourcetype=None, index=None):
        """
        Get all of the meta data belonging to a product of a given URN.

        mh: returns an iterator.
        """
        raise NotImplemented

    def meta(self, *args, **kwds):
        """
        Loads the meta-data info belonging to the product of specified URN.
        """
        return self.getMetaByUrn(*args, **kwds)

    def reference(self, ref):
        """
        Increment the reference count of a ProductRef.
        """
        # new ###
        poolname, dt, sn = parseUrn(ref.urn, int_index=True)
        _snd = self._dType[dt]['sn'][sn]
        if 'refcnt' not in _snd:
            _snd['refcnt'] = 0
        # assert self._urns[ref.urn]['refcnt'] == _snd['refcnt']
        _snd['refcnt'] += 1
        # /new ###
        if 0:
            if 'refcnt' not in self._urns:
                self._urns['refcnt'] = 0
            self._urns[ref.urn]['refcnt'] += 1

    def saveOne(self, prd, tag, geturnobjs, serialize_in, serialize_out, res, **kwds):
        """
        Save one product.

            # get the latest HK

            the new classes and tags.
               dTypes combines old classes and urns, is defined as::
                     $class_name0:
                            'currentSN': $sn
                            'sn':
                                $sn0:
                                   tags: [$tag1, $tag2, ...
                                   meta: [$start, $end]
                                   refcnt: $count
                                $sn1:
                                   tags: [$tag1, $tag2, ...
                                   meta: [$start, $end]
                                   refcnt: $count
               example::
                    'foo.bar.Bar':
                            'currentSN': 1
                            0:
                               'tags': ['cat', 'white']
                               'meta': [123, 456]
                               'refcnt': 0
                            1:
                               'tags': ['dog', 'white']
                               'meta': [321, 765]
                               'refcnt': 0
                    'foo.baz.Baz':
                            'currentSN': 34
                            34:
                               'tags': ['tree', 'green']
                               'meta': [100, 654]
                               'refcnt': 1
     `dTags` differs from `tag` 1. uses dType:[sn], instead of urn, so there is no poolname anywhere, 2. simplify by removing second level dict::
                     $tag_name0:
                           $class_name1:[$sn1, $sn2...]
                           $class_name2:[$sn3, ...]
               example::
                     'cat': { 'foo.bar.Bar':[0] }
                     'white': { 'foo.bar.Bar'; [0, 1] }
                     'dog': ...

        :res: list of results.
        :serialize_out: if True returns contents in serialized form.

        Return
        ------
        `list` of the following:
        `ProductRef`. `Urn` if `geturnobjs` is set. if`serialze_out` is set for `ProductRef` no product metadata is stored in the returned instance.
        The result is also stored in the `re` parameter.
        """
        if serialize_in:
            pn = fullname(prd)
            cls = prd.__class__
        else:
            # prd is json. extract prod name
            # '... "_STID": "Product"}]'
            pn = prd.rsplit('"', 2)[1]
            cls = Class_Look_Up[pn]
            pn = fullname(cls)
        with filelock.FileLock(self.lockpath('w')), \
                filelock.FileLock(self.lockpath('r')):
            # some new ####
            self._classes, self._tags, self._urns, \
                self._dTypes, self._dTags = tuple(
                    self.readHK().values())
            # /some new ####
            # if 0:
            #     c, t, u = self._classes, self._tags, self._urns

            #     if pn in c:
            #         sn = (c[pn]['currentSN'] + 1)
            #     else:
            #         sn = 0
            #         c[pn] = dict(sn=[])

            #     c[pn]['currentSN'] = sn
            #     c[pn]['sn'].append(sn)

            #     urn = makeUrn(poolname=self._poolname, typename=pn, index=sn)

            #     if urn not in u:
            #         u[urn] = {'tags': []}

            # new+old NORMALIZE TAGS###
            if tag is None:
                tags = []
            elif issubclass(tag.__class__, str):
                tags = [tag]
            elif issubclass(tag.__class__, list):
                tags = tag
            else:
                raise TypeError('Bad type for tag: %s.' %
                                tag.__class__.__name__)
            # new ####
            self._dTypes, self._dTags, _sn = \
                populate_pool2(tags, pn, sn=None,
                               dTypes=self._dTypes, dTags=self._dTags)

            # if 0:
            #     if tag is None:
            #         tags = []
            #     elif issubclass(tag.__class__, str):
            #         tags = [tag]
            #     elif issubclass(tag.__class__, list):
            #         tags = tag
            #     else:
            #         raise TypeError('Bad type for tag: %s.' %
            #                         tag.__class__.__name__)
            #     # new ####
            #     dTypes, dTags = self._dTypes, self._dTags

            #     if pn in dTypes:
            #         _sn = dTypes[pn]['currentSN'] + 1
            #     else:
            #         _sn = 0
            #         dTypes[pn] = {
            #             'currentSN': _sn,
            #             'sn': {}
            #         }

            #     snd = dTypes[pn]['sn']
            #     if _sn not in snd:
            #         snd[_sn] = {
            #             'tags': [],
            #             'meta': []
            #         }

            #     dTypes[pn]['currentSN'] = _sn
            #     urn = makeUrn(poolname=self._poolname, typename=pn, index=_sn)
            #     # /new #####

            #     for t in tags:
            #         self.setTag(t, urn)
            urn = makeUrn(poolname=self._poolname, typename=pn, index=_sn)
            try:
                # save prod and HK
                self.doSave(resourcetype=pn,
                            index=_sn,
                            data=prd,
                            tags=tags,
                            serialize_in=serialize_in,
                            serialize_out=serialize_out,
                            **kwds)
            except ValueError as e:
                msg = 'product ' + urn + ' saving failed.' + str(e) + trbk(e)
                logger.debug(msg)
                # some new ##
                self._classes, self._tags, self._urns, \
                    self._dTypes, self._dTags = tuple(
                        self.readHK().values())
                raise e

        if geturnobjs:
            if serialize_out:
                # return the URN string.
                res.append(urn)
            else:
                res.append(Urn(urn, poolurl=self._poolurl))
        else:
            rf = ProductRef(urn=Urn(urn, poolurl=self._poolurl),
                            poolmanager=self._poolmanager)
            if serialize_out:
                # return without meta
                res.append(rf)
            else:
                # it seems that there is no better way to set meta
                rf._meta = prd.getMeta()
                res.append(rf)

    def schematicSave(self, products, tag=None, geturnobjs=False, serialize_in=True, serialize_out=False, asyn=False, **kwds):
        """ do the scheme-specific saving.

        Parameters
        ----------
        product : BaseProduct, list
            Product or a list of them or '[ size1, prd, size2, prd2, ...]'.
        tag : str, list
            If given a tag, all products will be having this tag.
        If a list tags are given to every one product then the
        number of tags must not be the same to that of `product`. If
        they are equal, each tag is goven to the product at the same
        index in the `product` list.
        serialize_out : bool
            if `True` returns contents in serialized form.
        serialize_in : bool
            If set, product input is serialized.

        Returns
        -------
        ProductRef: Product reference.
        Urn: If `geturnobjs` is set.
        str: If `serialze_out` is set, serialized form of `ProductRef` or `URN`.
        list: `list` of the above of input is a list.
        """

        res = []
        alist = issubclass(products.__class__, list)
        json_list = False

        if alist:
            if isinstance(tag, list) and len(tag) != len(products):
                # make a list of tags to ','-separated tags
                tag = ','.join(t for t in tag if t)
            if isinstance(tag, str) or tag is None:
                tag = [tag] * len(products)

        if serialize_in:
            if not alist:
                prd = products
                self.saveOne(prd, tag, geturnobjs,
                             serialize_in, serialize_out, res, **kwds)
            else:
                if asyn:
                    prd = products
                    self.asyncSave(prd, tag, geturnobjs,
                                   serialize_in, serialize_out, res, **kwds)
                else:
                    for prd, t in zip(products, tag):
                        # result is in res
                        self.saveOne(prd, t, geturnobjs,
                                     serialize_in, serialize_out, res, **kwds)
        else:
            if alist:
                raise TypeError('a list cannot go with False serialize-in.')
            json_list = products.lstrip().startswith('[')
            if not json_list:
                prd = products
                self.saveOne(prd, tag, geturnobjs,
                             serialize_in, serialize_out, res, **kwds)
            else:
                # parse '[ size1, prd, size2, prd2, ...]'

                last_end = 1
                productlist = []
                comma = products.find(',', last_end)
                while comma > 0:
                    length = int(products[last_end: comma])
                    productlist.append(length)
                    last_end = comma + 1 + length
                    prd = products[comma + 2: last_end+1]
                    self.saveOne(prd, tag, geturnobjs,
                                 serialize_in, serialize_out, res, **kwds)
                    # +2 to skip the following ', '
                    last_end += 2
                    comma = products.find(',', last_end)
        if logger.isEnabledFor(logging_DEBUG):
            sz = 1 if not json_list and not alist else len(
                products) if serialize_in else len(productlist)
            logger.debug('%d product(s) generated %d %s: %s.' %
                         (sz, len(res), 'Urns ' if geturnobjs else 'prodRefs', lls(res, 200)))
        if alist or json_list:
            return serialize(res) if serialize_out else res
        else:
            return serialize(res[0]) if serialize_out else res[0]

    def doLoad(self, resourcetype, index, start=None, end=None, serialize_out=False):
        """ to be implemented by subclasses to do the action of loading
        """
        raise (NotImplementedError)

    def schematicLoad(self, resourcetype, index, start=None, end=None,
                      serialize_out=False):
        """ do the scheme-specific loading
        """

        with filelock.FileLock(self.lockpath('w')):
            ret = self.doLoad(resourcetype=resourcetype,
                              index=index, start=start, end=end,
                              serialize_out=serialize_out)
        return ret

    def doRemove(self, resourcetype, index, ayn=False):
        """ to be implemented by subclasses to do the action of reemoving
        """
        raise (NotImplementedError)

    def schematicRemove(self, urn=None, resourcetype=None, index=None, asyn=False, **kwds):
        """ do the scheme-specific removing
        """

        urn, datatype, sn = self.get_missing(
            urn, resourcetype, index, no_check=True)

        with filelock.FileLock(self.lockpath('w')),\
                filelock.FileLock(self.lockpath('r')):

            # get the latest HK
            # some new ####
            self._classes, self._tags, self._urns, \
                self._dTypes, self._dTags = tuple(
                    self.readHK().values())
            # c, t, u = self._classes, self._tags, self._urns
            # if urn not in u:
            #     raise ValueError(
            #         '%s not found in pool %s.' % (urn, self.getId()))
            datatypes, sns, alist = ProductPool.vectorize(datatype, sn)

            self.removeUrn(urn, datatype=datatype, sn=sn)

        if 0:
            n = max(lr, li)
            if lr and li and lr != li:
                raise TypeError(
                    f'Two args have different sizes {lr}, {li}.')
            alist = lr or li
            if n == 0:
                n = 1
            datatypes = datatype if lr else [datatype] * n
            sns = sn if li else [sn] * n

        res = self.doRemove(resourcetype=datatypes, index=sns, asyn=asyn)

        res1 = res if alist else [res]
        for i, r in enumerate(res1):
            if r is None:
                msg = f'product {urn[i]} removal failed.'
                if isinstance(self, (LocalPool, MemPool, HTTPClientpool)):
                    self._classes, self._tags, self._urns, \
                        self._dTypes, self._dTags = tuple(
                            self.readHK().values())
                    if getattr(self, 'ignore_error_when_delete', False):
                        raise
                    else:
                        logger.warning(msg)

                    # can only do one at a time
                    break
                elif isinstance(self, (PublicClientPool)):
                    self.getPoolInfo(update_hk=True)
        return res if alist else res[0]

    def doWipe(self):
        """ to be implemented by subclasses to do the action of wiping.
        """
        raise (NotImplementedError)

    def schematicWipe(self, asyn=False):
        """ do the scheme-specific wiping
        """
        with filelock.FileLock(self.lockpath('w')),\
                filelock.FileLock(self.lockpath('r')):
            # self._classes.clear()
            # self._tags.clear()
            # self._urns.clear()
            # new ##
            self._dTypes.clear()
            self._dTags.clear()
            # /new ##
            if asyn:
                res = self.doAsyncWipe()
            else:
                try:
                    res = self.doWipe()
                except ValueError as e:
                    msg = f'Wiping {self.poolname} failed. {e} traceback: {trbk(e)}'
                    if getattr(self, 'ignore_error_when_delete', False):
                        logger.warning(msg)
                    else:
                        raise
        return res

    def meta_filter(self, q, typename=None, reflist=None, urnlist=None, snlist=None, datatypes=None):
        """ returns filtered collection using the query.

        q is a MetaQuery
        valid inputs: typename and ns list; productref list; urn list; datatypes dict.

        :typename: data type (class name)
        :reflist: list of ProductRefs
        :urnlist: list of URNs
        :datatypes:  dict of {typename:sn_list}
        """

        ret = []
        qw = q.getWhere()

        if reflist:
            if isinstance(qw, str):
                code = compile(qw, 'qw.py', 'eval')
                for ref in reflist:
                    refmet = ref.getMeta()
                    m = refmet if refmet else self.getMetaByUrn(ref.urn)
                    if _eval(code=code, m=m):
                        ret.append(ref)
                return ret
            else:
                for ref in reflist:
                    refmet = ref.getMeta()
                    m = refmet if refmet else self.getMetaByUrn(ref.urn)
                    if qw(m):
                        ret.append(ref)
                return ret
        elif urnlist:
            if isinstance(qw, str):
                code = compile(qw, 'qw.py', 'eval')
                for urn in urnlist:
                    m = self.getMetaByUrn(urn)
                    if _eval(code=code, m=m):
                        ret.append(ProductRef(urn=urn, meta=m,
                                   poolmanager=self._poolmanager))
                return ret
            else:
                for urn in urnlist:
                    m = self.getMetaByUrn(urn)
                    if qw(m):
                        ret.append(ProductRef(urn=urn, meta=m,
                                   poolmanager=self._poolmanager))
                return ret
        elif snlist or datatypes:
            if isinstance(qw, str):
                code = compile(qw, 'qw.py', 'eval')
                if snlist:
                    datatypes = {typename: snlist}
                for cls in datatypes:
                    snlist = datatypes[cls]
                    for n in snlist:
                        urn = makeUrn(poolname=self._poolname,
                                      typename=typename, index=n)
                        m = self.getMetaByUrn(urn)
                        if _eval(code=code, m=m):
                            ret.append(ProductRef(urn=urn, meta=m,
                                       poolmanager=self._poolmanager))
                return ret
            else:
                if snlist:
                    datatypes = {typename: snlist}
                for cls in datatypes:
                    snlist = datatypes[cls]
                    for n in snlist:
                        urn = makeUrn(poolname=self._poolname,
                                      typename=typename, index=n)
                        m = self.getMetaByUrn(urn)
                        if qw(m):
                            ret.append(ProductRef(urn=urn, meta=m,
                                       poolmanager=self._poolmanager))
                return ret
        else:
            raise ('Must give a list of ProductRef or urn or sn')

    def prod_filter(self, q, cls=None, reflist=None, urnlist=None, snlist=None, datatypes=None):
        """ returns filtered collection using the query.

        q: an AbstractQuery.
        valid inputs: cls and ns list; productref list; urn list; datatypes dict.

        :cls: type. data type
        :reflist: list of ProductRefs
        :urnlist: list of URNs
        :datatypes:  dict of {cls:sn_list}
        """

        ret = []
        # will add query variable (e.g. 'p') to Global name space
        glbs = globals()
        qw = q.getWhere()
        var = q.getVariable()
        if var in glbs:
            savevar = glbs[var]
        else:
            savevar = 'not in glbs'

        if reflist:
            if isinstance(qw, str):
                code = compile(qw, 'qw.py', 'eval')
                for ref in reflist:
                    glbs[var] = pref.getProduct()
                    if _eval(code=code, m=m):
                        ret.append(ref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
            else:
                for ref in reflist:
                    glbs[var] = pref.getProduct()
                    if qw(m):
                        ret.append(ref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
        elif urnlist:
            if isinstance(qw, str):
                code = compile(qw, 'qw.py', 'eval')
                for urn in urnlist:
                    pref = ProductRef(urn=urn, poolmanager=self._poolmanager)
                    glbs[var] = pref.getProduct()
                    if _eval(code=code):
                        ret.append(pref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
            else:
                for urn in urnlist:
                    pref = ProductRef(urn=urn, poolmanager=self._poolmanager)
                    glbs[var] = pref.getProduct()
                    if qw(glbs[var]):
                        ret.append(pref)
                if savevar != 'not in glbs':
                    glbs[var] = savevar
                return ret
        elif snlist or datatypes:
            if isinstance(qw, str):
                code = compile(qw, 'qw.py', 'eval')
                if snlist:
                    datatypes = {cls.__name__: snlist}
                for typename in datatypes:
                    snlist = datatypes[typename]
                    cls = Class_Look_Up[typename.rsplit('.', 1)[-1]]
                    for n in snlist:
                        urno = Urn(cls=cls, poolname=self._poolname, index=n)
                        pref = ProductRef(
                            urn=urno, poolmanager=self._poolmanager)
                        glbs[var] = pref.getProduct()
                        if _eval(code=code):
                            ret.append(pref)
                    if savevar != 'not in glbs':
                        glbs[var] = savevar
                return ret
            else:
                if snlist:
                    datatypes = {cls.__name__: snlist}
                for typename in datatypes:
                    snlist = datatypes[typename]
                    cls = glbs[typename]
                    for n in snlist:
                        urno = Urn(cls=cls, poolname=self._poolname, index=n)
                        pref = ProductRef(
                            urn=urno, poolmanager=self._poolmanager)
                        glbs[var] = pref.getProduct()
                        if qw(glbs[var]):
                            ret.append(pref)
                    if savevar != 'not in glbs':
                        glbs[var] = savevar
                return ret
        else:
            raise ('Must give a list of ProductRef or urn or sn')

    def where(self, qw, prod='BaseProduct', urns=None):
        q = AbstractQuery(prod, 'p', qw)
        # if urns is None:
        # new ###
        datatypes = dict((k, list(v['sn'].keys()))
                         for k, v in self._dTypes.items())
        if 0:
            urns = self._urns.keys()
            res = self.prod_filter(q, prod, urnlist=urns)
        # new ###
        res2 = self.prod_filter(q, prod, datatypes=datatypes)

        # assert [r.urn for r in res] == [r.urn for r in res2]
        return [r.urn for r in res2]

    def doSelect(self, query, previous=None):
        """
        to be implemented by subclasses to do the action of querying.
        """
        raise (NotImplementedError)

    def schematicSelect(self,  query, previous=None):
        """
        do the scheme-specific querying.
        """
        is_MetaQ = issubclass(query.__class__, MetaQuery)
        is_AbstQ = issubclass(query.__class__, AbstractQuery)
        if not is_MetaQ and not is_AbstQ:
            raise TypeError('not a Query')
        lgb = Classes.mapping
        t, v, w, a = query.getType(), query.getVariable(
        ), query.getWhere(), query.retrieveAllVersions()
        ret = []
        if previous:
            this = (x for x in previous if x.urnobj.getPoolId()
                    == self._poolname)
            if is_MetaQ:
                ret += self.meta_filter(q=query, reflist=this)
            else:
                ret += self.prod_filter(q=query, reflist=this)
        else:
            # new ##
            # assert list(self._dTypes) == list(self._classes)
            for cname in self._dTypes:
                cls = lgb[cname.rsplit('.', 1)[-1]]
                if issubclass(cls, t):
                    # snlist = self._classes[cname]['sn']
                    # new ###
                    # assert snlist == list(self._dTypes[cname]['sn'])
                    snlist = list(self._dTypes[cname]['sn'])
                    if is_MetaQ:
                        ret += self.meta_filter(q=query, typename=cname,
                                                snlist=snlist)
                    else:
                        ret += self.prod_filter(q=query, cls=cls,
                                                snlist=snlist)

        return ret

    def __repr__(self):
        # co = ', '.join(str(k) + '=' + lls(v, 40)
        #               for k, v in self.__getstate__().items())
        co = ', '.join(str(k)+'=' + (v if issubclass(v.__class__, str) else
                                     f'< {v.__class__.__name__} {len(v)} >')
                       for k, v in self.__getstate__().items())
        return '<'+self.__class__.__name__ + ' ' + co + '>'

    def __getstate__(self):
        """ returns an odict that has all state info of this object.
        Subclasses should override this function.
        """
        return OrderedDict(
            poolname=getattr(self, '_poolname', 'unknown'),
            poolurl=getattr(self, '_poolurl', 'unknown'),
            _dTypes=self._dTypes,
            _dTags=self._dTags,
        )




# -*- coding: utf-8 -*-

""" https://livecodestream.dev/post/python-flask-api-starter-kit-and-project-layout/ """

from .route.getswag import swag

from .._version import __version__
from ..utils import getconfig
from ..utils.common import (getUidGid,
                            trbk,
                            logging_ERROR,
                            logging_WARNING,
                            logging_INFO,
                            logging_DEBUG, lls
                            )

from .session import init_session

from ..pal.poolmanager import PoolManager, DEFAULT_MEM_POOL

from flasgger import Swagger
from werkzeug.exceptions import HTTPException
from flask import Flask, make_response, jsonify, request
from werkzeug.routing import RequestRedirect
from werkzeug.routing import RoutingException, Map

import builtins
from os.path import expandvars
from weakref import WeakValueDictionary, getweakrefcount
import functools
from pathlib import Path
import sys
import json
import time
import os

# print(sys.path)
global logger

LOGGING_NORMAL = logging_INFO
""" routine logging level."""

LOGGING_DETAILED = logging_DEBUG
""" usually set to debugging """

cnt = 0
_BASEURL = ''


class PM_S(PoolManager):
    """Made to provid a different `_GlobalPoolList` useful for testing as a mock"""
    _GlobalPoolList = WeakValueDictionary()
    """ Another Global centralized dict that returns singleton -- the same -- pool for the same ID."""


def setup_logging(level=LOGGING_NORMAL, extras=None, tofile=None):
    import logging
    from logging.config import dictConfig
    from logging.handlers import QueueListener
    import queue
    que = queue.Queue(-1)  # no limit on size

    if extras is None:
        extras = LOGGING_NORMAL
    short = dict(format='%(asctime)s.%(msecs)03d'
                 ' %(levelname)4s'
                 ' %(filename)6s:%(lineno)3s'
                 ' - %(message)s',
                 datefmt="%y%m%d %H:%M:%S")
    detailed = dict(format='%(asctime)s.%(msecs)03d'
                    ' %(process)d %(thread)6d '
                    ' %(levelname)4s'
                    ' %(filename)6s:%(lineno)3s'
                    ' %(funcName)10s() - %(message)s',
                    datefmt="%Y%m%d %H:%M:%S")
    basedict = {
        'version': 1,
        'formatters': {'default': detailed, 'short': short},
        'handlers': {
            'wsgi': {
                'class': 'logging.StreamHandler',
                'stream': 'ext://flask.logging.wsgi_errors_stream',
                'formatter': 'default'
            },
            'non_block': {
                'class': 'logging.handlers.QueueHandler',
                # 'stream': 'ext://flask.logging.wsgi_errors_stream',
                'formatter': 'default',
                'queue': que,
            },

        },
        "loggers": {
            "werkzeug": {
                "level": LOGGING_NORMAL,
                "handlers": ["non_block"],
                "propagate": False
            }
        },
        'root': {
            'level': LOGGING_NORMAL,
            'handlers': ['wsgi']
        },
        'disable_existing_loggers': False
    }
    if tofile:
        basedict['handlers']['stream'] = {
            'class': 'logging.StreamHandler',
            'formatter': 'default',
            # level   : INFO
            # filters: [allow_foo]
            'stream': open(tofile, 'a')
        }
        basedict['root']['handlers'].append('stream')
    dict_config = dictConfig(basedict)

    if level < LOGGING_NORMAL:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(message)s"))
        logging_listener = QueueListener(
            que, handler, respect_handler_level=True)
        logging_listener.start()
    # logging.basicConfig(stream=sys.stdout, **detailed)
    # create logger
    if 1:
        for mod in ("requests", "filelock", ):
            logging.getLogger(mod).setLevel(extras)
        # logging.getLogger("werkzeug").setLevel(extras)
        if sys.version_info[0] > 2:
            logging.getLogger("urllib3").setLevel(extras)
    return logging

########################################
#### Config initialization Function ####
########################################


def init_conf_classes(pc, lggr):

    # setup user class mapping
    clp = pc['userclasses']
    lggr.debug('User class file '+clp)
    if clp == '':
        from ..dataset.classes import Classes as clz
        _bltn = dict((k, v) for k, v in vars(builtins).items() if k[0] != '_')
        clz.mapping.add_ns(_bltn, order=-1)
        return clz
    else:
        clpp, clpf = os.path.split(clp)
        sys.path.insert(0, os.path.abspath(clpp))
        # print(sys.path)
        # get the 'ProjectClasses' attribute
        projectclasses = __import__(clpf.rsplit('.py', 1)[0],
                                    globals(), locals(),
                                    ['ProjectClasses'], 0)
        clz = projectclasses.ProjectClasses
        lggr.debug('User classes: %d found.' % len(clz.mapping))
        return clz


SET_OWNER_MODE = False


@functools.lru_cache(6)
def checkpath(path, un):
    """ Checks  the directories and creats if missing.

    path: str. can be resolved with Path.
    un: server user name
    """
    # logger = current_app.logger

    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('path %s user %s' % (path, un))

    p = Path(path).resolve()
    if p.exists():
        if not p.is_dir():
            if logger.isEnabledFor(logging_ERROR):
                msg = str(p) + ' is not a directory.'
                logger.error(msg)
            return None
        else:
            # if path exists and can be set owner and group
            if p.owner() != un or p.group() != un:
                if logger.isEnabledFor(logging_WARNING):
                    msg = str(p) + ' owner %s group %s. Should be %s.' % \
                        (p.owner(), p.group(), un)
                    logger.warning(msg)
    else:
        # path does not exist

        if logger.isEnabledFor(logging_DEBUG):
            msg = str(p) + ' does not exist. Creating...'
            logger.debug(msg)
        p.mkdir(mode=0o775, parents=True, exist_ok=True)
        if logger.isEnabledFor(logging_INFO):
            logger.info(str(p) + ' directory has been made.')

    # logger.info('Setting owner, group, and mode...')
    if SET_OWNER_MODE and not setOwnerMode(p, un, logger):
        if logger.isEnabledFor(logging_INFO):
            logger.info('Cannot set owner %s to %s.' % (un, str(p)))
        return None

    return p


def setOwnerMode(p, username, logger):
    """ makes UID and GID set to those of self_username given in the config file. This function is usually done by the init script.
    """

    logger.debug('set owner, group to %s, mode to 0o775' % username)

    uid, gid = getUidGid(username)
    if uid == -1 or gid == -1:
        logger.debug(f'user {username} uid={uid} gid{gid}')
        return None
    try:
        os.chown(str(p), uid, gid)
        os.chmod(str(p), mode=0o775)
    except Exception as e:
        code, result, msg = excp(
            e,
            msg='cannot set dirs owner to ' +
            username + ' or mode. check config. ')
        logger.error(msg)
        return None

    return username


def init_httppool_server(app):
    """ Init a global HTTP POOL """

    # get settings from ~/.config/pnslocal.py config
    pc = app.config['PC']
    logger = app.logger
    # class namespace
    Classes = init_conf_classes(pc, logger)
    app.config['LOOKUP'] = Classes.mapping

    # client users
    from .model.user import getUsers
    app.config['USERS'] = getUsers(pc)

    # PoolManager is a singleton
    if PM_S.isLoaded(DEFAULT_MEM_POOL):
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug('cleanup DEFAULT_MEM_POOL')
        PM_S.getPool(DEFAULT_MEM_POOL).removeAll()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Done cleanup PoolManager.')
        logger.debug('ProcID %d. Got 1st request %s' %
                     (os.getpid(), str(app._got_first_request))
                     )
    PM_S.removeAll()

    # pool-related paths
    # the httppool that is local to the server
    scheme = 'server'
    _basepath = PM_S.PlacePaths[scheme]
    # this is SERVER_LOCAL_POOLPATH/data
    full_base_local_poolpath = os.path.join(_basepath, pc['api_version'])

    if checkpath(full_base_local_poolpath, pc['self_username']) is None:
        msg = 'Store path %s unavailable.' % full_base_local_poolpath
        logger.error(msg)
        return None

    app.config['POOLSCHEME'] = scheme

    # e.g. "/tmp/data/v0.13"
    app.config['FULL_BASE_LOCAL_POOLPATH'] = full_base_local_poolpath
    app.config['POOLURL_BASE'] = scheme + \
        '://' + full_base_local_poolpath + '/'


######################################
#### Application Factory Function ####
######################################


def create_app(config_object=None, level=None, logstream=None):
    """ If args have logger level, use it; else if debug is `True` set to 20
 use 'development' pnslocal.py config.

    :level: logging level. default `None`
    """
    config_object = config_object if config_object else getconfig.getConfig()

    global logger
    global _BASEURL

    _BASEURL = config_object['baseurl']
    _d = os.environ.get('PNS_DEBUG', None)
    _d = level if level else _d if _d else LOGGING_NORMAL

    if isinstance(_d, str):
        try:
            _d = int(_d)
        except TypeError:
            # must come from env -- ',' separated module list
            _d = _d.split(',')
    if isinstance(_d, list):
        debug_picked = _d
        level_picked = LOGGING_DETAILED
    else:
        # level number in str
        level_picked = _d
        debug_picked = []
    debug_picked.append('')
    logging = setup_logging(level=level_picked,
                            extras=int(config_object['logger_level_extras']),
                            tofile=logstream)
    logger = logging.getLogger('fdi.httppool_app')

    debug = (level_picked < logging_INFO)
    if 0 and debug_picked:
        for mod in debug_picked:
            mod = mod.strip()
            if not mod:
                continue
            if mod.startswith('='):
                logging.getLogger(mod[1:]).setLevel(logging_WARNING)
            else:
                logging.getLogger(mod).setLevel(level_picked)

        level = LOGGING_NORMAL
    else:
        level = level_picked
    logger.setLevel(level)
    # app = Flask('HttpPool', instance_relative_config=False,
    #            root_path=os.path.abspath(os.path.dirname(__file__)))
    app = Flask(__name__.split('.')[0], instance_relative_config=False,
                root_path=os.path.abspath(os.path.dirname(__file__)))
    app.logger = logger
    app.config_object = config_object
    app.config['LOGGER_LEVEL'] = level

    if os.environ.get('UW_DEBUG', False) in (1, '1', 'True', True):
        from remote_pdb import RemotePdb
        RemotePdb('127.0.0.1', 4444).set_trace()

    session = init_session(app)

    if debug:
        from werkzeug.debug import DebuggedApplication
        app.wsgi_app = DebuggedApplication(app.wsgi_app, True)
        app.debug = True
        logger.info('DEBUG mode %s' % (app.config['DEBUG']))
        app.config['PROPAGATE_EXCEPTIONS'] = True
    elif 'proxy_fix' in app.config_object:

        from werkzeug.middleware.proxy_fix import ProxyFix
        app.wsgi_app = ProxyFix(app.wsgi_app, **app.config_object['proxy_fix'])

    # from flask.logging import default_handler
    # app.logger.removeHandler(default_handler)

    app.config['SWAGGER'] = {
        'title': 'FDI %s HTTPpool Server' % __version__,
        'universion': 3,
        'openapi': '3.0.4',
        'specs_route': '/apidocs/',
        'url_prefix': _BASEURL
    }
    swag['servers'].insert(0, {
        'description': 'As in config file and server command line.',
        'url': config_object['scheme']+'://' +
        config_object['self_host'] + ':' +
        str(config_object['self_port']) +
        _BASEURL
    })
    swagger = Swagger(app, config=swag, merge=True)
    # swagger.config['specs'][0]['route'] = _BASEURL + s1
    app.config['PC'] = config_object

    # initialize_extensions(app)
    # register_blueprints(app)
    from .model.user import user
    app.register_blueprint(user, url_prefix=_BASEURL)

    from .route.pools import pools_api
    app.register_blueprint(pools_api, url_prefix=_BASEURL)
    from .route.httppool_server import data_api
    app.register_blueprint(data_api, url_prefix=_BASEURL)

    from .model.user import LOGIN_TMPLT
    if LOGIN_TMPLT:
        @app.errorhandler(401)
        @app.errorhandler(403)
        def handle_auth_error_codes(error):
            """ if verify_password returns False, this gets to run. """
            if error in [401, 403]:
                # send a login page
                if app.logger.isEnabledFor(logging_ERROR):
                    app.logger.error("Error %d. Start login page..." % error)
                page = make_response(render_template(LOGIN_TMPLT))
                return page
            else:
                raise ValueError('Must be 401 or 403. Nor %s' % str(error))

    # handlers for exceptions and some code
    add_errorhandlers(app)

    # Do not redirect a URL ends with no spash to URL/
    app.url_map.strict_slashes = False

    # with app.app_context():
    init_httppool_server(app)
    logger.info('Server initialized. logging level ' +
                str(app.logger.getEffectiveLevel()))

    @app.before_request
    def b4request():
        global cnt
        cnt += 1
        if logger.isEnabledFor(logging_DEBUG):
            args = request.view_args
            method = request.method
            logger.debug("%3d >>>[%4s] %s" %
                         (cnt, method, lls(str(args), 300)))
        elif logger.isEnabledFor(logging_INFO):
            # remove leading e.g. /fdi/v0.16
            s = request.path.split(_BASEURL)
            p = s[0] if len(s) == 1 else s[1] if s[0] == '' else request.path
            method = request.method
            logger.info("%3d >>>[%4s] %s" % (cnt, method, lls(p, 40)))

    return app


def add_errorhandlers(app):
    @app.errorhandler(Exception)
    def handle_excep(error):
        """ ref flask docs """
        ts = time.time()

        if issubclass(error.__class__, HTTPException) and error.code == 429:
            msg = "429 "
            error.code = 401
            response = make_response('', error)
        elif issubclass(error.__class__, HTTPException):
            if error.code == 409:
                spec = "Conflict or updating. "
            elif error.code == 500 and error.original_exception:
                error = error.original_exception
            else:
                spec = ''
            response = error.get_response()
            t = ' Traceback: ' + trbk(error)
            msg = '%s%d. %s, %s\n%s' % \
                (spec, error.code, error.name, error.description, t)
        elif issubclass(error.__class__, Exception):
            response = make_response()
            t = 'Traceback: ' + trbk(error)
            msg = '%s. %s.\n%s' % (error.__class__.__name__,
                                   str(error), t)
        else:
            response = make_response('', error)
            msg = ''
        w = {'result': 'FAILED', 'msg': msg, 'time': ts}
        response.data = json.dumps(w)
        response.content_type = 'application/json'
        return response



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from fdi.utils.getconfig import make_pool, getConfig

from fdi.pal.poolmanager import PoolManager, DEFAULT_MEM_POOL
from fdi.utils.common import trbk, fullname
from fdi.pal.context import Context, MapContext, RefContainer
from fdi.pal.productref import ProductRef
from fdi.pal.productstorage import ProductStorage
from fdi.pal.urn import Urn, parseUrn, parse_poolurl, makeUrn, UrnUtils
from fdi.pal.productpool import ProductPool
from fdi.pal.localpool import LocalPool
from fdi.pal.httpclientpool import HttpClientPool
from fdi.dataset.deserialize import deserialize
from fdi.pns.jsonio import commonheaders, auth_headers

from requests.auth import HTTPBasicAuth
import requests
import argparse
import os.path as op
import sys


def getPayload(aResponse, int_key=True):
    """ deserializes, if content_type is json, data or tex of responses from either the live server or the mock one.
    """

    x = aResponse.data if issubclass(
        aResponse.__class__, fwResponse) else aResponse.text
    if aResponse.headers['Content-Type'] == 'application/json':
        return deserialize(x, int_key=int_key)
    else:
        return x


rmk = '{"result": '
mmk = ', "msg": "'


def getresultjson(url, auth):
    x = requests.get(url, auth=auth)
    if not x.text.startswith(rmk):
        print(hkdata + ' must startswith '+rmk)
        sys.exit(1)
    r = x.text[len(rmk):]
    o = r.rsplit(mmk, 1)[0]
    # print(x.status_code, x.text)
    return o


def backup(pool, auth, fp0, tar):

    if tar:
        # make a backup tarfile
        tarf = clientpool.backup()
        fp = op.join(fp0, pool._poolname + '.tgz')
        print('backup to ', fp)
        with open(p, 'wb') as f:
            f.write(tarf)
        return
    # dump the contents using api
    print('backup to ', fp0)
    fullurl = pool._poolurl
    for hkdata in ['classes', 'tags', 'urns']:
        o = getresultjson(fullurl+'/hk/'+hkdata, auth)
        fp = op.join(fp0, hkdata + '.jsn')
        with open(fp, 'w') as hf:
            hf.write(o)

    urns = list(deserialize(o).keys())
    for u in urns:
        o = getresultjson(fullurl+'/'+u.split(':', 2)[-1], auth=auth)
        fp = op.join(fp0, '_'.join(u.rsplit(':', 2)[1:]))
        with open(fp, 'w') as hf:
            hf.write(o)


def restore(pool, auth, fp0, wipe=False):

    # restore from a backup tarfile
    fp = fp0
    with open(fp, 'rb') as f:
        tar = f.read()
    print('restore %s from ' % pool._poolurl, fp)
    lst = pool.restore(tar)
    # print(lst)
    return


if __name__ == '__main__':

    # default commandline options.
    pc = getConfig()
    verbose = False
    inputdir = None
    outputdir = None
    pool = ''
    onlyframes = -1
    wipe = False

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-u", "--poolurl", type=str,
                        default='', help='Backup from or restore to the pool by the pool name (used with config.py) or the pool URL given here. Default is not any pool.')
    parser.add_argument("-i", "--inputdir", type=str,
                        default=inputdir, help="Where to find saved data, relative to source dir. if not given, is resource/testdata of the package.")
    parser.add_argument("-o", "--outputdir", type=str,
                        default=outputdir, help='Output directory or poolname.')
    parser.add_argument('-U', '--username',
                        default=pc['username'], type=str, help='user name/ID')
    parser.add_argument('-P', '--password',
                        default=pc['password'], type=str, help='password')
    parser.add_argument("-n", "--number", type=int,
                        default=onlyframes, help='Only read this number of ?.')
    parser.add_argument("-t", "--tarfile",  action='store_true',
                        default=False, help="Backup to a gzipped tarfile by the server instead of getting every file in the pool.")
    parser.add_argument("-v", "--verbose", action='store_true',
                        default=verbose, help="Print more details.")

    args = parser.parse_args()
    pc['username'] = args.username
    pc['password'] = args.password
    verbose = args.verbose
    if args.verbose:
        print(args)
    if args.inputdir and args.outputdir:
        print(
            'You cannot backup (give "outputdir") and restore (give the "inputdir" at the same time.')
        sys.exit(-1)

    l = 0

    auth = HTTPBasicAuth(args.username, args.password)
    url = args.poolurl
    pstore = make_pool(url, auth=auth)
    clientpool = pstore.getPool(pstore.getPools(auth=auth, client=client)[0])
    print(clientpool)

    if args.outputdir is not None:
        backup(clientpool, auth, args.outputdir, args.tarfile)
    elif args.inputdir is None:
        print('Backup (give "outputdir") or restore (give the "inputdir").')
        sys.exit(-2)
    else:
        restore(clientpool, auth, args.inputdir)

# x = requests.get(url, auth=auth_headers)
# print(x.status_code, x.text)
